#!/usr/bin/env python3
"""
Examples:
    # 2010-12-15 setup genome schema in vervetdb.
    GenomeDB.py -v postgresql -k genome -d vervetdb -u yh
    
    #setup database in mysql
    #  (if it's already setup, output the pickled geneAnnotation data structure)
    GenomeDB.py -v mysql -z papaya -d genome -k "" -u yh -t 3702
        -o /tmp/geneAnnotationTax3702.pickle

Description:
    This is the genome database ORM.
"""
import sys, os
import logging
from sqlalchemy import DateTime, String, Integer, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy import UniqueConstraint
from datetime import datetime
from sqlalchemy import UniqueConstraint
from sqlalchemy import desc
from palos.utils import PassingData	
from palos.polymorphism.CNV import CNVSegmentBinarySearchTreeKey
from palos.algorithm.RBTree import RBDict
from palos.Genome import GeneModel
#2010-9-21 although "from Genome import GeneModel" works,
# it causes problem in pickle.load() because Genome is not directly visible outside.
from palos.db import Database, TableClass, get_sequence_segment

Base = declarative_base()
#Set it staticaly because SunsetDB is undefined at this point 
# and it has to be defined after this.
# 20200424 now that schema is set at the engine level.
#  Try not to use _schemaname_.
_schemaname_ = "genome_hg37"

class SequenceType(Base, TableClass):
    """
    2013.3.14 added description
    2008-07-27
        a table storing meta information to be referenced by other tables
    """
    __tablename__ = 'sequence_type'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    short_name = Column(String(223), unique=True)
    description = Column(Text)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class RawSequence(Base, TableClass):
    """
    A table to store chunks of sequences of entries from AnnotAssembly.
    annot_assembly_gi is not a foreign key anymore.
    annot_assembly_id replaces it.
    If the associated entry in AnnotAssembly is deleted,
        all raw sequences will be deleted as well.
    """
    __tablename__ = 'raw_sequence'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    annot_assembly_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.annot_assembly.id"))
    annot_assembly = relationship('AnnotAssembly')
    annot_assembly_gi = Column(Integer)
    start = Column(Integer)
    stop = Column(Integer)
    sequence = Column(String(10000))	#each fragment is 10kb
    UniqueConstraint('annot_assembly_id', 'start', 'stop', \
        name='raw_sequence_uniq_cnst')

class AnnotAssembly(Base, TableClass):
    """
    2011-8-24
        gi is no longer the primary key.
        a new id is added as primary key.
    2010-12-17
        Column raw_sequence_start_id is no longer a foreign key.
        To avoid circular dependency with RawSequence.
    2008-07-27
        table to store meta info of chromosome sequences
    """
    __tablename__ = 'annot_assembly'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    gi = Column(Integer)
    acc_ver = Column(String(32))
    accession = Column(String(32))
    version = Column(Integer)
    tax_id = Column(Integer)
    chromosome = Column(String(128))
    start = Column(Integer)
    stop = Column(Integer)
    genome_order = Column(Integer)
    orientation = Column(String(1))
    sequence = Column(String(10000))
    raw_sequence_start_id = Column(Integer)
    original_path = Column(Text)
    sequence_type_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.sequence_type.id"))
    sequence_type = relationship('SequenceType')
    chromosome_type_id = Column(Integer, 
        ForeignKey(f"{_schemaname_}.chromosome_type.id"))
    chromosome_type = relationship('ChromosomeType')
    genome_annotation_list = relationship('GenomeAnnotation',
        back_populates='annot_assembly',cascade='all,delete')
    #non-zero means outdated. to allow multiple outdated alignments
    outdated_index = Column(Integer, default=0)
    comment = Column(Text)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint(
        'accession', 'version','tax_id','chromosome', 'start', 'stop',
        'orientation', 'sequence_type_id', 'chromosome_type_id',
        'outdated_index')


class ChromosomeType(Base, TableClass):
    """
    store autosome, X, Y, mitochondrial
    """
    __tablename__ = 'chromosome_type'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)    
    short_name = Column(String(223), unique=True)
    comment = Column(Text)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class GenomeAnnotation(Base, TableClass):
    """
    2013.2.17
    """
    __tablename__ = 'genome_annotation'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    strand = Column(String(4))
    start = Column(Integer)
    stop = Column(Integer)
    description = Column(Text)
    comment = Column(Text)
    annot_assembly_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.annot_assembly.id"))
    annot_assembly = relationship('AnnotAssembly')
    genome_annotation_type_id = Column(Integer, \
        ForeignKey(f"{_schemaname_}.genome_annotation_type.id"))
    genome_annotation_type = relationship('GenomeAnnotationType')
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('strand', 'start', 'stop', 'annot_assembly_id',\
        'genome_annotation_type_id')


class GenomeAnnotationType(Base, TableClass):
    """
    2013.2.17
    table to store centromere, telemere, euchromatin, heterochromatin
    """
    __tablename__ = 'genome_annotation_type'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    short_name = Column(String(223), unique=True)
    description = Column(Text)
    comment = Column(Text)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class EntrezgeneType(Base, TableClass):
    """
    2008-07-28
        store the entrez gene types
    """
    __tablename__ = 'entrezgene_type'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    type = Column(String(223), unique=True)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class GeneCommentaryType(Base, TableClass):
    """
    2008-07-28
    """
    __tablename__ = 'gene_commentary_type'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    type = Column(String(223), unique=True)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class GeneCommentary(Base, TableClass):
    """
    2010-12-15
        gene linked to table Gene
    2008-07-28
        store different mRNAs/Peptides from the same gene or mRNA
    """
    __tablename__ = 'gene_commentary'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    accession = Column(String(32))
    version = Column(Integer)
    gi = Column(Integer)
    gene_id = Column(Integer, ForeignKey(f"{_schemaname_}.gene.id"))
    gene = relationship('Gene')
    gene_commentary_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene_commentary.id"))
    gene_commentary = relationship('GeneCommentary')
    #gene_commentary_ls = relationship('GeneCommentary',
    #   back_populates='gene_commentary', cascade='all,delete')
    start = Column(Integer)
    stop = Column(Integer)
    gene_commentary_type_id = Column(Integer, 
        ForeignKey(f"{_schemaname_}.gene_commentary_type.id"))
    gene_commentary_type = relationship('GeneCommentaryType')
    label = Column(Text)
    text = Column(Text)
    comment = Column(Text)
    gene_segment_ls = relationship('GeneSegment',
        back_populates='gene_commentary', cascade='all,delete')
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('gene_id', 'start', 'stop',
        'gene_commentary_type_id', 'gene_commentary_id')
    
    def getSequence(self, box_ls):
        """
2012.5.16
    it uses annot_assembly_id preferentially, if it is null,
        then uses genomic_gi.
2012.3.26
    get_sequence_segment()'s API has been changed.
    But this way of calling get_sequence_segment() is still outdated.
        valid only for banyan's mysql genome db
        because annot_assembly_id was not used.
        check GenomeDatabase.getSequenceSegment()
2009-01-03
        """
        seq = ''
        curs = __metadata__.bind
        #or self.table.bind or self.table.metadata.bind
        for box in box_ls:
            genomic_start, genomic_stop = box[:2]
            if self.gene.annot_assembly_id:
                annot_assembly_id = self.gene.annot_assembly_id
                gi = None
            elif self.gene.genomic_gi:
                gi = self.gene.genomic_gi
                annot_assembly_id = None
            else:
                gi = None
                annot_assembly_id = None
            schema = self._descriptor.table_options.get('schema')
            seq += get_sequence_segment(curs, gi=gi, start=genomic_start,
                stop=genomic_stop,
                annot_assembly_id=annot_assembly_id,
                annot_assembly_table='%s.%s'%(schema, AnnotAssembly.__tablename__),
                raw_sequence_table='%s.%s'%(schema, RawSequence.__tablename__))
        if self.gene.strand=='-1' and seq:	#need to reverse complement
            from Bio.Seq import Seq
            from Bio.Alphabet import IUPAC
            seq1 = Seq(seq, IUPAC.unambiguous_dna)
            seq2 = seq1.reverse_complement()
            seq = seq2.data
        return seq
    
    def getCDSsequence(self):
        """
        2009-01-03 implemented
        2008-09-22 implement later
        """
        if not getattr(self, 'protein_box_ls', None):
            self.construct_protein_box()
        
        self.cds_sequence = self.getSequence(self.protein_box_ls)
        #self.mrna_sequence = self.getSequence(self.mrna_box_ls)
    
    def construct_mrna_box(self):
        """
        2012.5.15
            GeneSegment.gene_commentary_type could be either exon or mRNA.
            add gene_segment.id into the mrna_box
            
        2010-8-18
            update to fetch exon/mRNA only because introns are now added in GeneSegment.
        2008-09-22
        """
        #either exon or mRNA type
        gene_commentary_type_id_ls = []
        gene_commentary_type = GeneCommentaryType.query.filter_by(type='exon').first()
        if gene_commentary_type:
            gene_commentary_type_id_ls.append(gene_commentary_type.id)
        gene_commentary_type = GeneCommentaryType.query.filter_by(type='mRNA').first()
        if gene_commentary_type:
            gene_commentary_type_id_ls.append(gene_commentary_type.id)
        
        gene_segment_ls = GeneSegment.query.\
            filter(GeneSegment.gene_commentary_type_id.in_(
                gene_commentary_type_id_ls)).\
            filter_by(gene_commentary_id=self.id)
        self.mrna_box_ls = []
        for gene_segment in gene_segment_ls:
            self.mrna_box_ls.append([gene_segment.start, gene_segment.stop,
                gene_segment.id])
        self.mrna_box_ls.sort()
        return self.mrna_box_ls
    
    def construct_protein_box(self):
        """
        2012.5.15
            GeneSegment.gene_commentary_type could be either CDS or peptide.
            add gene_segment.id into the protein_box
        2008-09-22
            find the corresponding protein gene_commentary if it's available.
            and construct a protein_box_ls
        """
        self.protein_box_ls = []
        if len(self.gene_commentary_ls)==1:
            gene_commentary = self.gene_commentary_ls[0]
        elif len(self.gene_commentary_ls)>1:
            logging.warn(f'More than 1 gene_commentary_ls for this '
                f'commentary id={self.id}, gene_id={self.gene_id}.')
            gene_commentary = self.gene_commentary_ls[0]
        else:
            gene_commentary = None
        if gene_commentary:
            #either exon or mRNA type
            gene_commentary_type_id_ls = []
            gene_commentary_type = GeneCommentaryType.query.filter_by(type='CDS').first()
            if gene_commentary_type:
                gene_commentary_type_id_ls.append(gene_commentary_type.id)
            gene_commentary_type = GeneCommentaryType.query.\
                filter_by(type='peptide').first()
            if gene_commentary_type:
                gene_commentary_type_id_ls.append(gene_commentary_type.id)
            
            gene_segment_ls = GeneSegment.query.\
                filter(GeneSegment.gene_commentary_type_id.in_(
                    gene_commentary_type_id_ls)).\
                filter_by(gene_commentary_id=gene_commentary.id)
            for gene_segment in gene_segment_ls:
                self.protein_box_ls.append([gene_segment.start, \
                    gene_segment.stop, gene_segment.id])
            self.protein_label = gene_commentary.label
            self.protein_comment = gene_commentary.comment
            self.protein_text = gene_commentary.text
        else:
            self.protein_label = None
            self.protein_comment = None
            self.protein_text = None
        self.protein_box_ls.sort()
        return self.protein_box_ls
    
    def constructAnnotatedBox(self):
        """
2010-8-18
    deal with the db (GeneSegment, etc.) populated by TAIRGeneXML2GenomeDB.py,
        not GeneASNXML2gene_mapping.py.
    GeneSegment store, UTR, CDS, exon, intron in their commentary_type already.
    
    #each entry is a tuple, (start, stop, box_type, is_translated, gene_segment.id, protein_box_index)
        """
        #each entry is a tuple, (start, stop, box_type, is_translated,
        #  gene_segment.id, protein_box_index)
        box_ls = []
        if len(self.gene_commentary_ls)==1:	#it's translated into protein.
            protein_commentary = self.gene_commentary_ls[0]
        elif len(self.gene_commentary_ls)>1:
            logging.warn(f'More than 1 gene_commentary_ls for this '
                f'commentary id={self.id}, gene_id={self.gene_id}.')
            protein_commentary = self.gene_commentary_ls[0]
        else:
            protein_commentary = None
        
        query = GeneSegment.query.filter_by(gene_commentary_id=self.id)
        if protein_commentary:
            #restrict the gene segment to intron only. 
            # exons will be covered by CDS and UTR.
            gene_commentary_type = GeneCommentaryType.query.\
                filter_by(type='intron').first()
            if gene_commentary_type:
                query = query.filter_by(
                    gene_commentary_type_id=gene_commentary_type.id)
        
        for gene_segment in query:
            box_ls.append([gene_segment.start, gene_segment.stop,\
                gene_segment.gene_commentary_type.type, 0,
                gene_segment.id, None])
        if protein_commentary:
            for gene_segment in protein_commentary.gene_segment_ls:
                if gene_segment.gene_commentary_type.type.find('UTR')!=-1:
                    is_translated = 0
                else:
                    is_translated = 1
                
                box_ls.append([gene_segment.start, gene_segment.stop,\
                    gene_segment.gene_commentary_type.type, \
                    is_translated, gene_segment.id,  None])
        box_ls.sort()
        return box_ls
        
    def construct_annotated_box(self):
        """
2012.5.15
    add gene_segment.id(or None) and cumulativeWithinCDSUTRAndIntronLen into box_ls
2008-10-01
    fix a bug that coordinates of a whole untranslated mrna block are
        replaced by that of a protein block
2008-10-01
    fix a bug that a whole untranslated mrna block got totally omitted.
2008-09-22
    combine mrna_box_ls and protein_box_ls to partition the whole gene into finer segments.
    box_ls = []	#each entry is a tuple,
        (start, stop, box_type, is_translated, protein_box_index,
            gene_segment.id, detailed_box_type, \
            cumulativeWithinCDSUTRAndIntronLen, exon_number)
    box_type = 'intron' or 'exon'. is_translated = 0 or 1. if it's translated,
        protein_box_index is index in protein_box_ls.
        """
        #each entry is a tuple, (start, stop, box_type, is_translated, protein_box_index)
        self.box_ls = []
        cumulativeWithinCDSUTRAndIntronLen=0
        #the cumulative length of all withinCDS UTRs and introns from the 5'\
        # till the designated CDS
        CDS_5_end_pos = -1	#5' end position of the whole CDS sequence
        if not hasattr(self, 'mrna_box_ls'):
            self.construct_mrna_box()
        
        if not hasattr(self, 'protein_box_ls'):
            self.construct_protein_box()
        
        no_of_mrna_boxes = len(self.mrna_box_ls)
        no_of_prot_boxes = len(self.protein_box_ls)
        if no_of_prot_boxes==0:
            # 2012.5.15 no protein, it's a non-coding gene
            default_detailed_box_type = 'exon'
        else:
            if self.gene.strand=='+1':
                default_detailed_box_type = '3UTR'
                #this is used after all proteins boxes
            else:
                default_detailed_box_type = '5UTR'
        j = 0	#index in protein_box_ls
        no_of_introns = 0
        exon_number = 0
        for i in range(no_of_mrna_boxes):
            mrna_start, mrna_stop, mrna_gene_segment_id = self.mrna_box_ls[i][:3]
            exon_number += 1
            if i>0:	#add intron if this is not the first exon
                intron_start = self.mrna_box_ls[i-1][1]+1	#stop of previous exon + 1
                intron_stop = mrna_start-1	#start of current exon + 1
                self.box_ls.append((intron_start, intron_stop, 'intron', 0, None, None, \
                    'intron', cumulativeWithinCDSUTRAndIntronLen, None))
                    #intron, no gene_segment
                no_of_introns += 1
                if CDS_5_end_pos!=-1:	#CDS has already begun
                    nonCDSLength = abs(intron_stop-intron_start)+1
                    cumulativeWithinCDSUTRAndIntronLen += nonCDSLength
            if j<no_of_prot_boxes:
                prot_start, prot_stop, prot_gene_segment_id = self.protein_box_ls[j][:3]
                if prot_start>=mrna_start and prot_stop<=mrna_stop:
                    if prot_start>mrna_start:	#one untranslated exon
                        if self.gene.strand=='+1':
                            detailed_box_type = '5UTR'
                        else:
                            detailed_box_type = '3UTR'
                        self.box_ls.append((mrna_start, prot_start-1, 'exon', 0, None, \
                            mrna_gene_segment_id, detailed_box_type, \
                            cumulativeWithinCDSUTRAndIntronLen, exon_number))
                            #UTR before CDS
                        if CDS_5_end_pos!=-1:	#CDS has already begun
                            nonCDSLength = abs(prot_start - mrna_start)
                            cumulativeWithinCDSUTRAndIntronLen += nonCDSLength
                    self.box_ls.append((prot_start, prot_stop, 'exon', 1, j, \
                        prot_gene_segment_id, 'CDS', 
                        cumulativeWithinCDSUTRAndIntronLen, exon_number))
                        #CDS
                    if CDS_5_end_pos==-1:	#first CDS
                        CDS_5_end_pos = prot_start
                    if prot_stop<mrna_stop:	#one more untranslated exon
                        if self.gene.strand=='+1':
                            detailed_box_type = '3UTR'
                        else:
                            detailed_box_type = '5UTR'
                        self.box_ls.append((prot_stop+1, mrna_stop, 'exon', 0, None, \
                            mrna_gene_segment_id, detailed_box_type, \
                            cumulativeWithinCDSUTRAndIntronLen, exon_number))
                        #UTR after CDS
                        if CDS_5_end_pos!=-1:	#CDS has already begun
                            nonCDSLength = abs(mrna_stop-prot_stop)
                            cumulativeWithinCDSUTRAndIntronLen += nonCDSLength
                    j += 1	#push protein box index up
                elif prot_stop<mrna_start:
                    #not supposed to happen
                    logging.error(f"protein box: [{prot_start}, {prot_stop}] of"
                        f" gene-id={self.gene_id} is ahead of mrna box "
                        f"[{mrna_start}, {mrna_stop}].")
                elif prot_start>mrna_stop:
                    if self.gene.strand=='+1':
                        detailed_box_type = '5UTR'
                    else:
                        detailed_box_type = '3UTR'
                    self.box_ls.append((mrna_start, mrna_stop, 'exon', 0, None, \
                        mrna_gene_segment_id, detailed_box_type, \
                        cumulativeWithinCDSUTRAndIntronLen, exon_number))
                    #2008-10-1 UTR
                    if CDS_5_end_pos!=-1:	#CDS has already begun
                            nonCDSLength = abs(mrna_stop-mrna_start+1)
                            cumulativeWithinCDSUTRAndIntronLen += nonCDSLength
                elif prot_start<=mrna_stop and prot_stop>mrna_stop:
                    #not supposed to happen
                    logging.error(f"protein box: [{prot_start}, {prot_stop}] of"
                        f" gene-id={self.gene_id} overlaps partially with "
                        f"mRNA box [{mrna_start}, {mrna_stop}].")
            else:	#passing all protein boxes
                self.box_ls.append((mrna_start, mrna_stop, 'exon', 0, None, \
                    mrna_gene_segment_id, default_detailed_box_type, \
                    cumulativeWithinCDSUTRAndIntronLen, exon_number))
                if CDS_5_end_pos!=-1:	#CDS has already begun
                    nonCDSLength = abs(mrna_stop-mrna_start+1)
                    cumulativeWithinCDSUTRAndIntronLen += nonCDSLength
                #UTR after all CDSs or exon
        self.CDS_5_end_pos = CDS_5_end_pos
        self.no_of_introns = no_of_introns
        return self.box_ls
    
    def outputSequenceInUpperLowerCase(self, outf):
        """
        2012.6.18
            get the box_ls from self.gene_segment_ls.
            If it's mRNA commentary, this function outputs pre-splicing mRNA.
            If it's protein commentary, this function outputs CDS sequence with introns.
        2012.6.11
            Segments present in GeneSegment are outputted in upper case.
            All inter-between segments are outputted in lower case.
        """
        """
        #2012.6.18 commented out
        import copy
        #use copy to avoid modifying self.mrna_box_ls and self.protein_box_ls
        box_ls = copy.deepcopy(self.construct_mrna_box())
        
        if not box_ls:	#this is a protein gene commentary.
            box_ls = copy.deepcopy(self.construct_protein_box())
        """
        box_ls = []
        for geneSegment in self.gene_segment_ls:
            box_ls.append([geneSegment.start, geneSegment.stop])
        box_ls.sort()
        
        outf.write(">GeneCommentary%s_fromGene%s\n"%(self.id, self.gene.id))
        if self.start<box_ls[0][0]:	#need to output the beginning in lower case.
            firstBoxStart = box_ls[0][0]
            sequence = self.getSequence([[self.start, firstBoxStart-1]])
            outf.write("%s"%(sequence.lower()))
        
        for i in range(len(box_ls)):
            box = box_ls[i]
            sequence = self.getSequence([[box[0], box[1]]])
            outf.write("%s"%(sequence.upper()))
            if i<len(box_ls)-1:
                #output the inter-box sequences in lower case
                next_box = box_ls[i+1]
                sequence = self.getSequence([[box[1]+1, next_box[0]-1]])
                outf.write("%s"%(sequence.lower()))
        if self.stop>box_ls[-1][1]:
            lastBoxStop = box_ls[-1][1]
            sequence = self.getSequence([[lastBoxStop+1, self.stop]])
            outf.write("%s"%(sequence.lower()))
        outf.write("\n")
    
    def getProteinSequence(self,):
        """
        2012.6.18
            assuming this is a mRNA gene commentary. won't work on its derivative protein commentary.
            get the protein_box_ls
            get all nucleotide sequence
            translate it into protein sequence
            return a fastq record.
        """
        self.getCDSsequence()
        #self.cds_sequence
        from Bio.Seq import Seq
        from Bio.Alphabet import IUPAC
        cds_seq = Seq(self.cds_sequence, IUPAC.unambiguous_dna)
        peptide = cds_seq.translate()
        return peptide
        
    
class GeneSegment(Base, TableClass):
    """
    2008-07-28
        table to store position info of segments (exon, intron, CDS, UTR ...)
         of gene-products (mRNA, peptide) in table GeneProduct
    """
    __tablename__ = 'gene_segment'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    gene_commentary_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene_commentary.id"))
    gene_commentary = relationship('GeneCommentary')
    gi = Column(Integer)
    start = Column(Integer)
    stop = Column(Integer)
    gene_commentary_type_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene_commentary_type.id"))
    gene_commentary_type = relationship('GeneCommentaryType')
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('gene_commentary_id', 'start', 'stop',
        'gene_commentary_type_id')

class Gene(Base, TableClass):
    """
    2012.4.26
        change the unique constraint to include annot_assembly_id,
         entrezgene_type_id and remove type_of_gene.
    2010-12-15
        move all EntrezgeneMapping-exclusive elements into Gene.
        EntrezgeneMapping will disappear.
    2008-07-27
        table to store meta info of genes
    """
    __tablename__ = 'gene'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    tax_id = Column(Integer)
    ncbi_gene_id = Column(Integer, unique=True)
    gene_symbol = Column(String(128))
    locustag = Column(String(128))
    synonyms = Column(Text)
    dbxrefs = Column(Text)
    chromosome = Column(String(115))
    strand = Column(String(4))
    start = Column(Integer)
    stop = Column(Integer)
    map_location = Column(String(256))
    description = Column(Text)
    type_of_gene = Column(String(128))
    symbol_from_nomenclature_authority = Column(String(256))
    full_name_from_nomenclature_authority = Column(Text)
    nomenclature_status = Column(String(64))
    other_designations = Column(Text)
    modification_date = Column(DateTime, default=datetime.now)
    genomic_accession = Column(String(32))
    genomic_version = Column(Integer)
    genomic_gi = Column(Integer)
    annot_assembly_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.annot_assembly.id"))
    annot_assembly = relationship('AnnotAssembly')
    entrezgene_type_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.entrezgene_type.id"))
    entrezgene_type = relationship('EntrezgeneType')
    comment = Column(Text)
    gene_commentary_ls = relationship('GeneCommentary',
        back_populates='gene',cascade='all,delete')
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('tax_id', 'locustag', 'chromosome', 'strand', 'start',
        'stop', 'entrezgene_type_id', 'annot_assembly_id')

class Gene2go(Base, TableClass):
    """
    2010-12-15
        add go_qualifier as part of the unique constraint
        "NOT" means the gene is not associated with specified GO.
    2008-07-27
        table to store mapping between gene and GO
    """
    __tablename__ = 'gene2go'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    tax_id = Column(Integer)
    gene_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene.id"))
    gene = relationship('Gene')
    go_id = Column(String(32))
    evidence = Column(String(3))
    go_qualifier = Column(String(64))
    go_description = Column(Text)
    pubmed_ids = Column(Text)
    category = Column(String(16))
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('tax_id', 'gene_id', 'go_id', 'evidence', 'category',
        'go_qualifier')

class Gene2Family(Base, TableClass):
    """
    2010-8-19 a table recording which family TE belongs to
    """
    __tablename__ = 'gene2family'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    gene_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene.id"))
    gene = relationship('Gene')
    family_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene_family.id"))
    family = relationship('GeneFamily')
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('gene_id', 'family_id')


class GeneFamily(Base, TableClass):
    """
    2010-8-19
        family names and etc.
    """
    __tablename__ = 'gene_family'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    short_name = Column(String(212), unique=True)
    parent_family_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene_family.id"))
    parent_family = relationship('GeneFamily')
    family_type = Column(String(256))
    description = Column(Text)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class README(Base, TableClass):
    __tablename__ = 'readme'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    title = Column(Text)
    description = Column(Text)
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)

class Gene_symbol2id(Base, TableClass):
    """
    2010-6-21
        add created_by, updated_by, date_created, date_updated
    2008-07-27
        a derived table from Gene in order to map all available
         gene names (symbol, locustag, synonym) to gene-id
    """
    __tablename__ = 'gene_symbol2id'
    __table_args__ = {'schema':_schemaname_}
    id = Column(Integer, primary_key=True)
    tax_id = Column(Integer)
    gene_symbol = Column(String(256))
    gene_id = Column(Integer,
        ForeignKey(f"{_schemaname_}.gene.id"))
    gene = relationship('Gene')
    symbol_type = Column(String(64))
    created_by = Column(String(256))
    updated_by = Column(String(256))
    date_created = Column(DateTime, default=datetime.now)
    date_updated = Column(DateTime)
    UniqueConstraint('gene_id', 'gene_symbol', 'symbol_type')

def getEntrezgeneAnnotatedAnchor(db, tax_id):
    """
    2011-1-25
        row.gene_id -> row.id
    2008-08-13
        similar to annot.bin.codense.common.get_entrezgene_annotated_anchor, 
        but use elixir db interface
    """
    sys.stderr.write("Getting entrezgene_annotated_anchor ...")
    chromosome2anchor_gene_tuple_ls = {}
    gene_id2coord = {}
    offset_index = 0
    block_size = 5000
    rows = db.queryTable(Gene).filter_by(tax_id=tax_id).offset(offset_index).\
        limit(block_size)
    while rows.count()!=0:
        for row in rows:
            genomic_gi = row.genomic_gi
            chromosome = row.chromosome
            gene_id = row.id
            strand = row.strand
            start = row.start
            stop = row.stop
            if strand=='1' or strand=='+1' or strand=='+':
                strand = '+'
            elif strand=='-1' or strand=='-':
                strand = '-'
            else:
                strand = strand
            
            if chromosome not in chromosome2anchor_gene_tuple_ls:
                chromosome2anchor_gene_tuple_ls[chromosome] = []
            
            chromosome2anchor_gene_tuple_ls[chromosome].append((start, gene_id))
            gene_id2coord[gene_id] = (start, stop, strand, genomic_gi)
        
        rows = db.queryTable(Gene).filter_by(tax_id=tax_id).\
            offset(offset_index).limit(block_size)
    for chromosome in chromosome2anchor_gene_tuple_ls:
        #sort the list
        chromosome2anchor_gene_tuple_ls[chromosome].sort()
    sys.stderr.write("Done.\n")
    return chromosome2anchor_gene_tuple_ls, gene_id2coord

class OneGenomeData(PassingData):
    """
    chrOrder (order of chromosomes)
        1: order by column genome_order, ascendingly
        2: order by size, descendingly
        3: order by column genome_order, ascendingly
    
    2011-11-28 add two arguments
        self.chrOrder = None
        self.sequence_type_id = None
    2011-3-25
        a structure to wrap data related to one genome
         (identified by tax_id)
        
        All these data used to be handled by class GenomeDatabase.
            Previous solution is not good because GenomeDatabase
            is an umbrella of genome data from different species.
    """
    def __init__(self, db_genome=None, tax_id=3702, maxPseudoChrSize=3000000000, **keywords):
        """
        2012.8.13
            add argument maxPseudoChrSize, 
            which determines the max pseudo chromosome size (for self.chr_id2cumu_chr_start)
        #2011-3-25
        """
        #PassingData.__init__(self)
        self.db_genome = db_genome
        self._chr_id2size = None
        self._chr_id2cumu_size = None
        self._chr_id2cumu_start = None
        self._chr_id2cumu_chr_start = None
        #2012.8.13 the key is chromosome ID.
        # the value is (pseudoChromosomeID, cumuStart).
        # This data structure is used to re-organize the old chromosomes into new/pseudo chromosomes.
        # The max length of each pseudo chromosome is <=self.maxPseudoChrSize. 
        
        #2013.2.17
        self._chr_id2centromere = None
        self._chr_id2annot_assembly = None
        
        self.tax_id = tax_id
        self.maxPseudoChrSize = maxPseudoChrSize
        
        self.chr_gap = None
        self._chr_id_ls = None
        #2013.07.29 changed from chr_id_ls to _chr_id_ls 
        
        self.chrOrder = None
        self.sequence_type_id = None
        self._cumuSpan2ChrRBDict = None
        
        PassingData.__init__(self, **keywords)
        #keywords could contain chr_gap, chrOrder
    
    @property
    def chr_id_ls(self):
        """
        2013.07.29 return or construct a list of chromosome IDs according to user's chrOrder
        """
        if self._chr_id_ls is None:
            self.chr_id_ls = (self.tax_id, self.chrOrder, )
        return self._chr_id_ls
    
    @chr_id_ls.setter
    def chr_id_ls(self, argument_ls=[]):
        """
        #2014.01.12 added outdated_index=0 to filter AnnotAssembly
        2013.07.29
        2014.01.12 set outdated_index to 0
        """
        if argument_ls and len(argument_ls)>0:
            tax_id = argument_ls[0]
            if len(argument_ls)>1:
                chrOrder = argument_ls[1]
            else:
                chrOrder = self.chrOrder
        else:
            tax_id = self.tax_id
            chrOrder = self.chrOrder
        sys.stderr.write("Getting chr_id_ls with tax_id=%s, chrOrder=%s ..."%(tax_id, chrOrder))
        
        query = self.db_genome.queryTable(AnnotAssembly).filter_by(tax_id=tax_id).\
            filter_by(start=1).filter_by(outdated_index=0)
        if self.chrOrder==2:
            query = query.order_by(AnnotAssembly.stop)
        else:	#1 or 3
            query = query.order_by(AnnotAssembly.genome_order)
        if self.sequence_type_id:
            query = query.filter_by(sequence_type_id=self.sequence_type_id)
        
        self._chr_id_ls = []
        
        for row in query:
            self._chr_id_ls.append(row.chromosome)
        sys.stderr.write("%s chromosomes.\n"%(len(self._chr_id_ls)))
    
    @property
    def chr_id2size(self):
        """
        2011-3-13
        """
        if self._chr_id2size is None:
            self.chr_id2size = (self.tax_id, )
        return self._chr_id2size
    
    @chr_id2size.setter
    def chr_id2size(self, argument_ls=[]):
        """
        #2014.01.12 added outdated_index=0 to filter AnnotAssembly
        #2011-11-21
            order the chromosomes according to self.chrOrder
            select sequences based on self.sequence_type_id
        2011-3-12
            modified from get_chr_id2size() of variation/src/common.py
            keywords could include, tax_id=3702.
        2008-10-07 curs could be elixirdb.metadata.bind
        2007-10-12
        """
        if argument_ls and len(argument_ls)>0:
            tax_id = argument_ls[0]
        else:
            tax_id = self.tax_id
        sys.stderr.write("Getting chr_id2size for tax_id %s ..."%(tax_id))
        
        #query = AnnotAssembly.query.filter_by(tax_id=tax_id).filter_by(start=1)
        """
        if self.chrOrder==2:
            orderByString = 'order by stop desc'
        else:
            orderByString = ""
        if self.sequence_type_id:
            extraCondition = " and sequence_type_id=%s "%self.sequence_type_id
        else:
            extraCondition = ""
        
        query = self.db_genome.metadata.bind.execute(
            "select id, chromosome, stop from genome.%s where tax_id=%s and start=1 %s %s"%\
            (AnnotAssembly.table.name, tax_id, extraCondition, orderByString))
        """
        query = self.db_genome.queryTable(AnnotAssembly).filter_by(tax_id=tax_id).\
            filter_by(start=1).filter_by(outdated_index=0)
        if self.chrOrder==2:
            query = query.order_by(AnnotAssembly.stop)
        else:	#1 or 3
            query = query.order_by(AnnotAssembly.genome_order)
        if self.sequence_type_id:
            query = query.filter_by(sequence_type_id=self.sequence_type_id)
        
        self._chr_id2size = {}
        self._chr_id2annot_assembly = {}
        
        for row in query:
            self._chr_id2annot_assembly[row.chromosome] = row
            chr_id = row.chromosome
            size = row.stop
            self._chr_id2size[chr_id] = size
        #self._chr_id2size = chr_id2size
        sys.stderr.write("%s chromosomes.\n"%(len(self._chr_id2size)))
    
    @property
    def chr_id2cumu_size(self):
        """
        2011-3-13
        """
        if self._chr_id2cumu_size is None:
            self.chr_id2cumu_size = (self.tax_id, self.chr_gap)
        return self._chr_id2cumu_size
    
    @chr_id2cumu_size.setter
    def chr_id2cumu_size(self, argument_list):
        """
        2011-3-12
            modified from get_chr_id2cumu_size() of variation/src/common.py
            cumu_size of one chr = sum of length of
             (all prior chromosomes + current chromosome) + all gaps between them
        2008-02-04
            add _chr_id_ls
            turn chr_id all into 'str' form
        2008-02-01
            add chr_gap, copied from variation.src.misc
        2007-10-16
        """
        tax_id, chr_gap = argument_list[:2]
        sys.stderr.write("Getting chr_id2cumu_size for %s ..."%tax_id)
        if self._chr_id2size is None:
            self.chr_id2size = (tax_id,)
        #if chr_gap not specified, take the 1/5th of the average chromosome size as its value
        if chr_gap==None:
            chr_size_ls = self.chr_id2size.values()
            chr_gap = int(sum(chr_size_ls)/(5.0*len(chr_size_ls)))
        
        if self._chr_id_ls is None:
            self.chr_id_ls = (tax_id, self.chrOrder, )
        
        #_chr_id_ls = sorted(self.chr_id2size)
        #no more chromosome 0
        first_chr = self.chr_id_ls[0] 
        #chr_id_ls might not be continuous integers. so dictionary is better
        self._chr_id2cumu_size = {first_chr:self.chr_id2size[first_chr]}
        for i in range(1,len(self.chr_id_ls)):
            chr_id = self.chr_id_ls[i]
            prev_chr_id = self.chr_id_ls[i-1]
            self._chr_id2cumu_size[chr_id] = \
                self._chr_id2cumu_size[prev_chr_id] + chr_gap + self.chr_id2size[chr_id]
        sys.stderr.write("%s chromosomes.\n"%(len(self._chr_id2cumu_size)))
    
    @property
    def chr_id2cumu_start(self):
        """
        2011-3-13
        """
        if self._chr_id2cumu_start is None:
            self.chr_id2cumu_start = (self.tax_id, self.chr_gap)
        return self._chr_id2cumu_start
    
    @chr_id2cumu_start.setter
    def chr_id2cumu_start(self, argument_list):
        """
        chrOrder
            =1: alphabetical order
            =2: by size , descendingly
        
        2011-11-21
            deal with chrOrder
                =1: alphabetical order
                =2: by size , descendingly
        2011-4-22
            cumu_start is now 0-based, which makes it easy to generate adjusted coordinates.
                new_start = cumu_start + start.
        2011-3-15
            Treat one genome of multiple chromsomes as one continuous chromosome.
            
            For one chr, cumu_start = sum of length of all prior chromosomes + all gaps between them.
            
            cumu_start is 0-based.
            
            The difference between chr_id2cumu_size and chr_id2cumu_start is 
                that the former includes the length of
                the current chromosome and the gap before it.
        """
        tax_id, chr_gap = argument_list[:2]
        sys.stderr.write("Getting chr_id2cumu_start for %s, ..."%tax_id)
        if self._chr_id2size is None:
            self.chr_id2size = [tax_id]
        #if chr_gap not specified, take the 1/5th of the average chromosome size as its value
        if chr_gap==None:
            chr_size_ls = self.chr_id2size.values()
            chr_gap = int(sum(chr_size_ls)/(5.0*len(chr_size_ls)))
        
        if self._chr_id_ls is None:
            self.chr_id_ls = (tax_id, self.chrOrder, )
        
        #chr_id_ls = sorted(self.chr_id2size)
        """
        if self.chrOrder==2:
            size_chr_id_ls = [(value, key) for key, value in self.chr_id2size.items()]
            size_chr_id_ls.sort()
            size_chr_id_ls.reverse()
            chr_id_ls = [row[1] for row in size_chr_id_ls]
        """
        first_chr = self.chr_id_ls[0] 
        self._chr_id2cumu_start = {first_chr:0}
        #chr_id_ls might not be continuous integers. so dictionary is better
        #start from 0.
        for i in range(1, len(self.chr_id_ls)):
            chr_id = self.chr_id_ls[i]
            prev_chr_id = self.chr_id_ls[i-1]
            self._chr_id2cumu_start[chr_id] = \
                self._chr_id2cumu_start[prev_chr_id] + chr_gap + self.chr_id2size[prev_chr_id]
        sys.stderr.write("%s chromosomes.\n"%(len(self._chr_id2cumu_start)))
    
    @property
    def chr_id2cumu_chr_start(self):
        """
        2012.8.13
        """
        if self._chr_id2cumu_chr_start is None:
            self.chr_id2cumu_chr_start = (self.maxPseudoChrSize, self.tax_id, self.chr_gap)
        return self._chr_id2cumu_chr_start
    
    @chr_id2cumu_chr_start.setter
    def chr_id2cumu_chr_start(self, argument_list):
        """
        2012.8.13
        this structure is slightly different from chr_id2cumu_start.
            chr_id2cumu_start is for grouping all chromosomes into one giant chromosomes.
            chr_id2cumu_chr_start is for groupping all chromosomes into several chromosomes,
                size of each is <= self.maxPseudoChrSize. 
        """
        
        maxPseudoChrSize, tax_id, chr_gap = argument_list[:3]
        sys.stderr.write("Getting chr_id2cumu_chr_start for tax-id=%s, "
            " chr-gap=%s, maxPseudoChrSize=%s ..."%(
                tax_id, chr_gap, maxPseudoChrSize))
        if self._chr_id2size is None:
            self.chr_id2size = (tax_id, )
        #if chr_gap not specified, take the 1/5th of the average chromosome size as its value
        if chr_gap==None:
            chr_size_ls = self.chr_id2size.values()
            chr_gap = int(sum(chr_size_ls)/(5.0*len(chr_size_ls)))
        
        if self._chr_id_ls is None:
            self.chr_id_ls = (tax_id, self.chrOrder, )
        """
        chr_id_ls = sorted(self.chr_id2size)
        if self.chrOrder==2:
            size_chr_id_ls = [(value, key) for key, value in self.chr_id2size.items()]
            size_chr_id_ls.sort()
            size_chr_id_ls.reverse()
            chr_id_ls = [row[1] for row in size_chr_id_ls]
        """
        first_chr = self.chr_id_ls[0]
        newChrID = 1
        self._chr_id2cumu_chr_start = {first_chr:(newChrID, 0)}
            #chr_id_ls might not be continuous integers. so dictionary is better
            #start from 0.
        for i in range(1, len(self.chr_id_ls)):
            chr_id = self.chr_id_ls[i]
            prev_chr_id = self.chr_id_ls[i-1]
            prev_new_chr, prev_cumu_start = self._chr_id2cumu_chr_start[prev_chr_id]
            new_cumu_start = prev_cumu_start + chr_gap + self.chr_id2size[prev_chr_id]
            if new_cumu_start>=self.maxPseudoChrSize:
                newChrID += 1	#start a new fake chromosome
                new_cumu_start =0	#reset
            else:
                newChrID = prev_new_chr
            self._chr_id2cumu_chr_start[chr_id] = (newChrID, new_cumu_start)
        sys.stderr.write("%s chromosomes, last newChrID=%s.\n"%(len(self._chr_id2cumu_chr_start), newChrID))
    
    @property
    def chr_id2centromere(self):
        """
        2013.2.17
        """
        if self._chr_id2centromere is None:
            self.chr_id2centromere = (self.maxPseudoChrSize, self.tax_id, self.chr_gap)
        return self._chr_id2centromere
    
    @chr_id2centromere.setter
    def chr_id2centromere(self, argument_list):
        """
        2013.2.17
        """
        
        maxPseudoChrSize, tax_id, chr_gap = argument_list[:3]
        sys.stderr.write("Getting chr_id2centromere for tax-id=%s, chr-gap=%s, maxPseudoChrSize=%s ..."%\
                        (tax_id, chr_gap, maxPseudoChrSize))
        if self._chr_id2cumu_start is None:
            self.chr_id2cumu_start = (self.tax_id, self.chr_gap)
        
        self._chr_id2centromere ={}
        for chr_id, annot_assembly in self._chr_id2annot_assembly.items():
            for genome_annotation in annot_assembly.genome_annotation_list:
                if genome_annotation.genome_annotation_type_id==1:
                    self._chr_id2centromere[chr_id] = genome_annotation
        sys.stderr.write("%s centromeres.\n"%(len(self._chr_id2centromere)))
    
    @property
    def cumuSpan2ChrRBDict(self):
        """
        2011-3-25
        """
        if self._cumuSpan2ChrRBDict is None:
            self.cumuSpan2ChrRBDict = (self.tax_id, self.chr_gap)
        return self._cumuSpan2ChrRBDict
    
    @cumuSpan2ChrRBDict.setter
    def cumuSpan2ChrRBDict(self, argument_list):
        """
        2011-4-22
            adjust because chr_id2cumu_start is now 0-based.
            the position in cumuSpan2ChrRBDict is 1-based.
        2011-3-25
            turn it into a setter of cumuSpan2ChrRBDict
            usage example:
                tax_id = 3702
                chr_gap = 0
                self.cumuSpan2ChrRBDict = (tax_id, chr_gap)
            
        2011-3-16
            Treat one genome of multiple chromsomes as one continuous chromosome (uber-chomosome).
            This function creates a RB dictionary, which holds a map between 
                uberchromosome coordinate and individual chromosome coordinate.
        """
        tax_id, chr_gap = argument_list[:2]
        sys.stderr.write("Creating cumuSpan2ChrRBDict for tax_id=%s, chr_gap=%s \n"%(
            tax_id, chr_gap))
        self._cumuSpan2ChrRBDict = RBDict()
        if self._chr_id2size is None:
            self.chr_id2size = [tax_id,]
        if self._chr_id2cumu_start is None:
            self.chr_id2cumu_start = (tax_id, chr_gap)
        for chr_id, cumu_start in self.chr_id2cumu_start.items():
            chr_size = self.chr_id2size.get(chr_id)
            span_ls=[cumu_start+1, cumu_start+chr_size]
            segmentKey = CNVSegmentBinarySearchTreeKey(
                chromosome=0,
                span_ls=span_ls,
                min_reciprocal_overlap=0.00000000000001,)
                #2010-8-17 overlapping keys are regarded as separate instances
                #  as long as they are not identical.
            if segmentKey not in self._cumuSpan2ChrRBDict:
                self._cumuSpan2ChrRBDict[segmentKey] = [chr_id, 1, chr_size]
            else:
                logging.error("%s of chr %s is already in cumuSpan2ChrRBDict."%\
                    (segmentKey, chr_id))
        sys.stderr.write("%s chromosomes.\n"%(len(self._cumuSpan2ChrRBDict)))

class GenomeDatabase(Database):
    __doc__ = __doc__
    option_default_dict = Database.option_default_dict.copy()
    option_default_dict[('drivername', 1,)][0] = 'postgresql'
    option_default_dict.update({
        ('geneAnnotationPickleFname', 0, ): ['', 'o', 1, 
            'The optional output file to contain a pickled object with three attributes: \n'
            '1. gene_id2model \n'
            '2. chr_id2gene_id_ls \n'
            '3. geneSpanRBDict. \n'
            'OPTIONAL: only used when you run GenomeDB.py as a standalone program'
            ],
        ('tax_id', 0, int): [3702, 't', 1, 'Taxonomy ID for geneAnnotationPickleFname.'],\
        })
    def __init__(self, **keywords):
        """
        wrap _chr_id2size, _chr_id2cumu_size, _chr_id2cumu_start etc. into OneGenomeData class.
        """
        Database.__init__(self, **keywords)
        #2011-3-25
        self.tax_id2genomeData = {}
    
    
    def get_gene_id2model(self, tax_id=3702, needSequence=False):
        """
        argument needSequence: whether CDS_sequence,
            mrna_sequence is fetched for each gene_commentary
        2010-10-11
            bug fix: several genes could have the same segmentKey in geneSpanRBDict.
        2009-1-3
            add cds_sequence, mrna_sequence to new_gene_commentary
        2008-10-01
            sort EntrezgeneMapping by chromosome, start
        2008-10-1
            similar to variation.src.GenomeBrowser.get_gene_id2model()
             but adapts to the new genome db schema
            construct a data structure to hold whole-genome annotation of genes
        """
        sys.stderr.write("Getting gene_id2model and chr_id2gene_id_ls ...\n")
        gene_id2model = {}
        chr_id2gene_id_ls = {}
        geneSpanRBDict = RBDict()
        i = 0
        block_size = 50000
        query = self.queryTable(Gene).filter_by(tax_id=tax_id).\
            order_by(Gene.chromosome).order_by(Gene.start)
        rows = query.offset(i).limit(block_size)
        while rows.count()!=0:
            for row in rows:
                chromosome = row.chromosome
                gene_id = row.id
                
                if chromosome not in chr_id2gene_id_ls:
                    chr_id2gene_id_ls[chromosome] = []
                chr_id2gene_id_ls[chromosome].append(gene_id)
                if gene_id not in gene_id2model:
                    gene_id2model[gene_id] = GeneModel(gene_id=gene_id, 
                        ncbi_gene_id=row.ncbi_gene_id, chromosome=chromosome,
                        gene_symbol=row.gene_symbol,\
                        locustag=row.locustag, map_location=row.map_location,\
                        type_of_gene=row.entrezgene_type.type,
                        type_id=row.entrezgene_type_id,\
                        start=row.start, stop=row.stop, strand=row.strand,
                        tax_id=row.tax_id,
                        description =row.description)
                        #2010-8-19 add description
                    
                    gene_model = gene_id2model[gene_id]
                    if gene_model.chromosome and gene_model.start and gene_model.stop:
                        segmentKey = CNVSegmentBinarySearchTreeKey(
                            chromosome=gene_model.chromosome, \
                            span_ls=[gene_model.start, gene_model.stop], \
                            min_reciprocal_overlap=1, strand=gene_model.strand,
                            gene_id=gene_model.gene_id,
                            gene_start=gene_model.start,
                            gene_stop=gene_model.stop)
                        #2010-8-17 any overlap short of being identical is tolerated.
                        if segmentKey not in geneSpanRBDict:
                            geneSpanRBDict[segmentKey] = []
                        geneSpanRBDict[segmentKey].append(gene_id)
                
                    for gene_commentary in row.gene_commentary_ls:
                        if not gene_commentary.gene_commentary_id:
                            #ignore gene_commentary that are derived from other gene_commentary_ls.
                            #  they'll be handled within the parental gene_commentary.
                            #gene_commentary.constructAnnotatedBox()
                            gene_commentary.construct_annotated_box()
                            #2010-9-21 it sets protein_label and etc.
                            #  constructAnnotatedBox() doesn't do this.
                            if needSequence:
                                gene_commentary.getCDSsequence()
                            #pass everything to a database-independent object, easy for pickling
                            new_gene_commentary = GeneModel(gene_id=gene_id, \
                                gene_commentary_id=gene_commentary.id, \
                                start=gene_commentary.start, stop=gene_commentary.stop, \
                                label=gene_commentary.label, comment=gene_commentary.comment, \
                                text=gene_commentary.text,\
                                gene_commentary_type=gene_commentary.gene_commentary_type.type,\
                                protein_label=gene_commentary.protein_label,\
                                protein_comment=gene_commentary.protein_comment,\
                                protein_text=gene_commentary.protein_text,\
                                mrna_box_ls=gene_commentary.mrna_box_ls,\
                                protein_box_ls=gene_commentary.protein_box_ls,\
                                box_ls=gene_commentary.box_ls,
                                cds_sequence = getattr(gene_commentary, 'cds_sequence', None),
                                mrna_sequence = getattr(gene_commentary, 'mrna_sequence', None))
                            gene_id2model[gene_id].gene_commentary_ls.append(new_gene_commentary)
                else:
                    logging.error(f"gene {gene_id} already exists in gene_id2model.")
                i += 1
            if self.report:
                sys.stderr.write("%s\t%s\t"%('\x08'*80, i))
            rows = query.offset(i).limit(block_size)
        
        sys.stderr.write("Done.\n")
        return gene_id2model, chr_id2gene_id_ls, geneSpanRBDict
    
    def createGenomeRBDict(self, tax_id=3702, max_distance=20000, debug=False):
        """
        2012.5.15
            use gene_commentary.construct_annotated_box() instead of
                gene_commentary.constructAnnotatedBox()
            one more step toward deprecating get_gene_id2model()
        2012.3.19
            add an attribute to geneSpanRBDict:
                genomeRBDict.genePadding (=max_distance)
        2011-3-24
            stop casting row.chromosome into integer (just string type)
            add ncbi_gene_id to oneGeneData
        2011-3-10
            copied from FindCNVContext.py
        2011-1-27
            require Gene.start, Gene.stop, Gene.chromosome not null.
            Gene.id is the new gene_id (was Gene.gene_id).
        2010-10-3
            bug fixed: (chr, start, stop) is not unique.
            There are genes with the same coordinates.
        """
        sys.stderr.write("Creating a RBDict for all genes from organism %s ... \n"%tax_id)
        genomeRBDict = RBDict()
        genomeRBDict.genePadding = max_distance
        query = self.queryTable(Gene).filter_by(tax_id=tax_id).filter(Gene.start!=None).\
            filter(Gene.stop!=None).filter(Gene.chromosome!=None)
        counter = 0
        real_counter = 0
        for row in query:
            #try:	# convert to integer except when "C" or "M"/mitochondria is encountered.
            #	chromosome = int(row.chromosome)
            # #integer chromosomes should be converted as CNV.chromosome is integer.
            #except:
            chromosome = row.chromosome
            segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=chromosome, \
                span_ls=[max(1, row.start - max_distance), row.stop + max_distance], \
                min_reciprocal_overlap=1,)
                #2010-8-17 overlapping keys are regarded as separate instances as long as they are not identical.
            if segmentKey not in genomeRBDict:
                genomeRBDict[segmentKey] = []
            oneGeneData = PassingData(strand = row.strand, gene_id = row.id, \
                gene_start = row.start, \
                gene_stop = row.stop, geneCommentaryRBDictLs=[],\
                ncbi_gene_id=row.ncbi_gene_id, type_of_gene=row.type_of_gene)
            counter += 1
            for gene_commentary in row.gene_commentary_ls:
                if not gene_commentary.gene_commentary_id:
                    # ignore gene_commentary that are derived from other gene_commentary_ls. 
                    # they'll be handled within the parental gene_commentary.
                    geneCommentaryRBDict = RBDict()
                    geneCommentaryRBDict.gene_commentary_id = gene_commentary.id
                    box_ls = gene_commentary.construct_annotated_box()
                    geneCommentaryRBDict.box_ls = box_ls
                    geneCommentaryRBDict.protein_box_ls= gene_commentary.protein_box_ls
                    gene_commentary.getCDSsequence()
                    geneCommentaryRBDict.cds_sequence = gene_commentary.cds_sequence
                    geneCommentaryRBDict.gene_commentary_type_name = gene_commentary.gene_commentary_type.type
                    geneCommentaryRBDict.start = gene_commentary.start
                    geneCommentaryRBDict.stop = gene_commentary.stop
                    geneCommentaryRBDict.CDS_5_end_pos = gene_commentary.CDS_5_end_pos
                    geneCommentaryRBDict.no_of_introns = gene_commentary.no_of_introns
                    
                    #box_ls = gene_commentary.constructAnnotatedBox()
                    #2012.5.15 constructAnnotatedBox()
                    #  requires GeneSegment to have UTR,exon,CDS,intron already in place
                    #box_ls = gene_commentary.box_ls
                    no_of_boxes = len(box_ls)
                    
                    numberPorter = PassingData(cds_number = 0,\
                                            intron_number = 0,\
                                            utr_number = 0,\
                                            exon_number = 0)
                    for i in range(no_of_boxes):
                        if row.strand == "-1":	#reverse
                            box = box_ls[-i-1]
                        else:
                            box = box_ls[i]
                        start, stop, box_type, is_translated, protein_box_index, \
                            gene_segment_id, detailed_box_type, \
                            cumulativeWithinCDSUTRAndIntronLen, exon_number = box[:9]
                        
                        geneSegmentKey = CNVSegmentBinarySearchTreeKey(
                            chromosome=chromosome, 
                            span_ls=[start, stop], \
                            min_reciprocal_overlap=1, label=detailed_box_type,
                            is_translated=is_translated,
                            protein_box_index=protein_box_index,utr_number=None,\
                            cds_number=None, intron_number=None, exon_number=None, \
                            gene_segment_id=gene_segment_id, 
                            cumulativeWithinCDSUTRAndIntronLen=cumulativeWithinCDSUTRAndIntronLen)
                        #2010-8-17 overlapping keys are regarded as
                        #  separate instances as long as they are not identical.
                        
                        if detailed_box_type=='3UTR' or detailed_box_type=='5UTR' or detailed_box_type=='UTR':
                            numberPorter.utr_number += 1
                            geneSegmentKey.utr_number = numberPorter.utr_number
                        elif detailed_box_type=='CDS':
                            numberPorter.cds_number += 1
                            geneSegmentKey.cds_number = numberPorter.cds_number
                        elif detailed_box_type=='intron':
                            numberPorter.intron_number += 1
                            geneSegmentKey.intron_number = numberPorter.intron_number
                        if box_type=='exon':
                            #detailed_box_type could be exon only for non-coding genes. 
                            geneSegmentKey.exon_number = exon_number
                        
                        geneCommentaryRBDict[geneSegmentKey] = None
                        real_counter += 1
                    oneGeneData.geneCommentaryRBDictLs.append(geneCommentaryRBDict)
            genomeRBDict[segmentKey].append(oneGeneData)
            if counter%1000==0:
                sys.stderr.write("%s%s\t%s"%('\x08'*100, counter, real_counter))
                if debug:
                    break
        sys.stderr.write("%s%s\t%s\n"%('\x08'*100, counter, real_counter))
        sys.stderr.write("%s Done.\n"%(str(genomeRBDict)))
        return genomeRBDict
    
    def dealWithGenomeRBDict(self, genomeRBDictPickleFname, tax_id=3702, \
        max_distance=20000, debug=None):
        """
        2011-3-10
        """
        sys.stderr.write("Dealing with genomeRBDict ...")
        import pickle
        if genomeRBDictPickleFname:
            if os.path.isfile(genomeRBDictPickleFname):
                #if this file is already there, suggest to un-pickle it.
                picklef = open(genomeRBDictPickleFname)
                genomeRBDict = pickle.load(picklef)
                del picklef
            else:
                #if the file doesn't exist, but the filename is given,
                #  pickle snps_context_wrapper into it
                genomeRBDict = self.createGenomeRBDict(tax_id=tax_id, \
                    max_distance=max_distance, debug=debug)
                #2008-09-07 pickle the snps_context_wrapper object
                picklef = open(genomeRBDictPickleFname, 'w')
                pickle.dump(genomeRBDict, picklef, -1)
                picklef.close()
        else:
            genomeRBDict = self.createGenomeRBDict(tax_id=tax_id, \
                max_distance=max_distance, debug=debug)
        sys.stderr.write("%s unique genomic spans. Done.\n"%(len(genomeRBDict)))
        return genomeRBDict
    
    def getOneGenomeData(self, tax_id=3702, chr_gap=0, chrOrder=1, \
        sequence_type_id=None, \
        maxPseudoChrSize=3000000000, **keywords):
        """
        chrOrder (order of chromosomes)
            1: order by column genome_order, ascendingly
            2: order by size, descendingly
            3: order by column genome_order, ascendingly
        
        2012.8.13
            add argument maxPseudoChrSize, keywords
        2011-11-21
            add argument chrOrder (order of chromosomes)
                1: whatever order in database
                2: order by size, descendingly
        2011-3-25
            API to get & set OneGenomeData for one taxonomy.
        """
        if tax_id not in self.tax_id2genomeData:
            oneGenomeData = OneGenomeData(db_genome=self, tax_id=tax_id, \
                chr_gap=chr_gap, chrOrder=chrOrder,\
                sequence_type_id=sequence_type_id,
                maxPseudoChrSize=maxPseudoChrSize, **keywords)
            self.tax_id2genomeData[tax_id] = oneGenomeData
        return self.tax_id2genomeData.get(tax_id)
    
    def outputGenomeSequence(self, tax_id=None, sequence_type_id=1, \
        fastaTitlePrefix='chr', outputDir=None, chunkSize=70):
        """
        #2014.01.12 added outdated_index=0 to filter AnnotAssembly
        2011-7-7
at this moment, The output is with symap in mind.

i.e.:
outputDir = os.path.expanduser("~/script/variation/bin/symap_3_3/data/pseudo/vervet177BAC/sequence/pseudo/")
db_genome.outputGenomeSequence(tax_id=60711, sequence_type_id=10,
    fastaTitlePrefix='BAC', outputDir=outputDir, chunkSize=70)
        """
        sys.stderr.write("Outputting genome sequences from tax %s, seq-type %s to %s ...\n"\
            %(tax_id, sequence_type_id, outputDir))
        query = self.queryTable(AnnotAssembly).filter_by(tax_id=tax_id).\
            filter_by(sequence_type_id=sequence_type_id).\
            filter_by(outdated_index=0).order_by(AnnotAssembly.id)
        for row in query:
            fastaTitle = '%s%s'%(fastaTitlePrefix, row.id)
            outputFname = os.path.join(outputDir, '%s.seq'%(fastaTitle))
            outf = open(outputFname, 'w')
            outf.write(">%s\n"%(fastaTitle))
            sub_query = self.queryTable(RawSequence).filter_by(annot_assembly_id=row.id).\
                order_by(RawSequence.start)
            for seq_row in sub_query:
                sequence = seq_row.sequence
                while len(sequence)>0:
                    outf.write("%s\n"%(sequence[:chunkSize]))
                    sequence = sequence[chunkSize:]
            del outf
        sys.stderr.write("Done.\n")
    
    def getSequenceSegment(self, tax_id=None, chromosome=None, start=None,
        stop=None, schema='genome'):
        """
        #2014.01.12 added outdated_index=0 to filter AnnotAssembly
        2011-10-24
            start and stop are 1-based.
        """
        chromosome = str(chromosome)
        query = self.queryTable(AnnotAssembly).filter_by(tax_id=tax_id).\
            filter_by(chromosome=chromosome).filter_by(outdated_index=0).\
            order_by(AnnotAssembly.id)
        annot_assembly_id_ls = []
        for row in query:
            annot_assembly_id_ls.append(row.id)
        if len(annot_assembly_id_ls)>1:
            logging.warn("more than 1 entries from tax_id %s, chromosome %s. take first one."\
                %(tax_id, chromosome))
        elif len(annot_assembly_id_ls)==0:
            logging.warn("No entry available from tax_id %s, chromosome %s."%(\
                tax_id, chromosome))
            return ""
        annot_assembly_id = annot_assembly_id_ls[0]
        curs = __metadata__.bind
        #or self.table.bind or self.table.metadata.bind
        seq = get_sequence_segment(curs, gi=None, start=start, stop=stop, \
            annot_assembly_id=annot_assembly_id,\
            annot_assembly_table='%s.%s'%(schema, AnnotAssembly.__tablename__), \
            raw_sequence_table='%s.%s'%(schema, RawSequence.__tablename__))
        return seq
    
    def getSequenceType(self, short_name=None, entry_id=None, **keywords):
        """
        2013.3.14
        """
        db_entry = self.checkIfEntryInTable(TableClass=SequenceType, 
            short_name=short_name, entry_id=entry_id)
        if not db_entry:
            db_entry = SequenceType(short_name=short_name)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def getTopNumberOfChomosomes(self, contigMaxRankBySize=100, \
        contigMinRankBySize=1, tax_id=60711, sequence_type_id=9,\
        version=None, chromosome_type_id=0, chromosome=None, outdated_index=0):
        """
        #2014.01.12 added outdated_index=0 to filter AnnotAssembly
        2013.3.19 added argument outdated_index
        2013.3.14 added argument version, chromosome
        2013.2.15 added argument chromosome_type_id
        chromosome_type_id: what type of chromosomes to be included,
             same as table genome.chromosome_type.
            0: all, 1: autosomes, 2: X, 3:Y, 4: mitochondrial
        2011-11-7
            copied from vervet/src/AlignmentToCallPipeline.py
        2011-11-6
            rename argument topNumberOfContigs to contigMaxRankBySize
            add argument contigMinRankBySize
        2011-9-13
            return chr2size instead of a set of ref names
        2011-7-12
            get all the top contigs
        """
        no_of_contigs_to_fetch = contigMaxRankBySize-contigMinRankBySize+1
        print(f"Getting {no_of_contigs_to_fetch} chromosomes with rank "
            f"(by size) between {contigMinRankBySize} and {contigMaxRankBySize}:",
            flush=True)
        chr2size = {}
        query = self.queryTable(AnnotAssembly).filter_by(tax_id=tax_id).\
            filter_by(sequence_type_id=sequence_type_id).\
            filter_by(outdated_index=0)
        if chromosome_type_id:
            query = query.filter_by(chromosome_type_id=chromosome_type_id)
        if version is not None:
            query = query.filter_by(version=version)
        if chromosome:
            query = query.filter_by(chromosome=chromosome)
        if outdated_index is not None:
            query = query.filter_by(outdated_index=outdated_index)
        #order by size from big to small
        query = query.order_by(desc(AnnotAssembly.stop))
        counter = 0
        for row in query:
            counter += 1
            if counter>=contigMinRankBySize and counter<=contigMaxRankBySize:
                chr2size[row.chromosome] = row.stop
            if len(chr2size)>=no_of_contigs_to_fetch:
                break
        print(f"{len(chr2size)} chromosomes.", flush=True)
        return chr2size
    
    def getChromosomeType(self, short_name=None, id=None, comment=None,
        **keywords):
        """
        2013.3.14
        """
        TableClass=ChromosomeType
        db_entry = self.checkIfEntryInTable(TableClass=TableClass, \
            short_name=short_name, entry_id=id)
        if not db_entry:
            db_entry = TableClass(short_name=short_name, comment=comment)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def checkAnnotAssembly(self, id=None, accession = None, \
        version=None, tax_id=None, chromosome =None, \
        start=None, stop =None, orientation=None, \
        sequence_type_id=None, chromosome_type_id=None,\
        outdated_index=0):
        """
        2013.3.15 added argument outdated_index
        2013.3.14
            all arguments part of unique key
        """
        if id is not None:
            db_entry = self.checkIfEntryInTable(
                TableClass=AnnotAssembly, entry_id=id)
            if db_entry:
                return db_entry
        
        query = self.queryTable(AnnotAssembly)
        if accession:
            query = query.filter_by(accession=accession)
        if version:
            query = query.filter_by(version=version)
        if tax_id:
            query = query.filter_by(tax_id=tax_id)
        if chromosome:
            query = query.filter_by(chromosome=chromosome)
        if start:
            query = query.filter_by(start=start)
        if stop:
            query = query.filter_by(stop=stop)
        if sequence_type_id:
            query = query.filter_by(sequence_type_id=sequence_type_id)
        if chromosome_type_id:
            query = query.filter_by(chromosome_type_id=chromosome_type_id)
        if outdated_index:
            query = query.filter_by(outdated_index=outdated_index)
        db_entry = query.first()
        return db_entry
    
    def getAnnotAssembly(self, id=None, gi=None, acc_ver=None,
        accession = None, version =None, tax_id=None, chromosome =None,
        start =1, stop =None, orientation=None, sequence = None,\
        raw_sequence_start_id=None, original_path=None, \
        sequence_type_id=None, sequence_type_name= None, \
        chromosome_type_id=None, chromosome_type_name=None, \
        outdated_index=0,\
        comment=None, **keywords):
        """
        2013.3.15 added argument outdated_index
        2013.3.14 more fully-fledged
        2012.4.26
            return db entry of a chromosome
        """
        if not sequence_type_id:
            if sequence_type_name:
                sequence_type = self.getSequenceType(short_name=sequence_type_name)
                sequence_type_id=sequence_type.id
        if not chromosome_type_id:
            if chromosome_type_name:
                chromosome_type = self.getChromosomeType(short_name=chromosome_type_name)
                chromosome_type_id=chromosome_type.id
        db_entry=self.checkAnnotAssembly(id=id, accession=accession, \
            version=version, tax_id=tax_id, \
            chromosome=chromosome, start=start, stop=stop,
            orientation=orientation,
            sequence_type_id=sequence_type_id,
            chromosome_type_id=chromosome_type_id,
            outdated_index=outdated_index)
        if not db_entry:
            db_entry = AnnotAssembly(gi =gi, acc_ver=acc_ver, accession=accession,
                version=version, tax_id = tax_id, chromosome = chromosome,
                start = start, stop =stop, orientation=orientation,
                sequence=sequence, raw_sequence_start_id = raw_sequence_start_id,
                original_path=original_path,
                sequence_type_id = sequence_type_id,
                chromosome_type_id=chromosome_type_id,
                outdated_index=outdated_index,\
                comment = comment)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def checkGenomeAnnotation(self, entry_id=None, strand=None, start=None,
        stop =None, annot_assembly_id=None,
        genome_annotation_type_id=None, description=None, comment=None):
        """
        2013.08.01
        """
        if entry_id is not None:
            db_entry = self.checkIfEntryInTable(
                TableClass=GenomeAnnotation, entry_id=entry_id)
            if db_entry:
                return db_entry
        
        query = self.queryTable(GenomeAnnotation)
        if strand:
            query = query.filter_by(strand=strand)
        if start:
            query = query.filter_by(start=start)
        if stop:
            query = query.filter_by(stop=stop)
        if annot_assembly_id:
            query = query.filter_by(annot_assembly_id=annot_assembly_id)
        if genome_annotation_type_id:
            query = query.filter_by(
                genome_annotation_type_id=genome_annotation_type_id)
        db_entry = query.first()
        return db_entry
    
    def getGenomeAnnotation(self, entry_id=None, strand=None,
        start=None, stop =None,
        annot_assembly_id=None, \
        tax_id=None, chromosome=None, version=None, \
        sequence_type_name=None, sequence_type_id=None, \
        genome_annotation_type_id=None, genome_annotation_type_name=None,
        description=None, comment=None, **keywords):
        """
        2013.8.1
        """
        if not annot_assembly_id:
            if tax_id and chromosome:
                annot_assembly = self.getAnnotAssembly(id=None, gi=None,
                    acc_ver=None, accession=None, version=version,
                    tax_id=tax_id, chromosome=chromosome, start=None, stop=None,
                    orientation=None, sequence=None, raw_sequence_start_id=None,
                    original_path=None, sequence_type_id=sequence_type_id,
                    sequence_type_name=sequence_type_name,
                    chromosome_type_id=None,
                    chromosome_type_name=None, outdated_index=None,
                    comment=None)
                annot_assembly_id = annot_assembly.id
        if not genome_annotation_type_id:
            if genome_annotation_type_name:
                genome_annotation_type = self.getGenomeAnnotationType(
                    short_name=genome_annotation_type_name, description=None,
                    comment=None)
                genome_annotation_type_id = genome_annotation_type.id
        db_entry=self.checkGenomeAnnotation(entry_id=entry_id, strand=strand,
            start=start, stop=stop,
            annot_assembly_id=annot_assembly_id, \
            genome_annotation_type_id=genome_annotation_type_id, \
            description=description, comment=comment)
        if not db_entry:
            db_entry = GenomeAnnotation(strand = strand, start = start,\
                stop =stop, annot_assembly_id=annot_assembly_id, \
                genome_annotation_type_id = genome_annotation_type_id, \
                description=description, comment = comment)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def checkGenomeAnnotationType(self, entry_id=None, short_name=None):
        """
        2013.08.01
        """
        if id is not None:
            db_entry = self.checkIfEntryInTable(
                TableClass=GenomeAnnotationType, entry_id=entry_id)
            if db_entry:
                return db_entry
        
        query = self.queryTable(GenomeAnnotationType)
        if short_name:
            query = query.filter_by(short_name=short_name)
        db_entry = query.first()
        return db_entry
    
    def getGenomeAnnotationType(self, entry_id=None, short_name=None, \
        description=None, comment=None, **keywords):
        """
        2013.8.1
        """
        db_entry=self.checkGenomeAnnotationType(entry_id=entry_id,
            short_name=short_name)
        if not db_entry:
            db_entry = GenomeAnnotationType(short_name = short_name, \
                description=description, comment = comment)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def getGeneCommentary(self, gene_id=None, start=None, stop=None,
        gene_commentary_type_id=None, gene_commentary_id=None,\
        accession=None, version=None, gi=None, label=None, text=None,
        comment=None,
        gene=None, gene_commentary=None):
        """
        2012.4.26
        """
        if gene_id is None and gene.id is not None:
            gene_id = gene.id
        if gene_commentary_id is None and gene_commentary is not None:
            gene_commentary_id = gene_commentary.id
        
        if gene_id and start and stop and gene_commentary_type_id:
            query = self.queryTable(GeneCommentary)
            query = query.filter_by(gene_id=gene_id).filter_by(start=start).\
                filter_by(stop=stop).\
                filter_by(gene_commentary_type_id=gene_commentary_type_id)
            if gene_commentary_id:
                query = query.filter_by(gene_commentary_id=gene_commentary_id)
            geneCommentary = query.first()
        else:
            geneCommentary = None
        if not geneCommentary:
            geneCommentary = GeneCommentary(start=start, stop=stop, \
                gene_id=gene_id, gene_commentary_type_id=gene_commentary_type_id,
                gene_commentary_id=gene_commentary_id, accession=accession,
                version=version,\
                gi=gi, label=label,text=text, comment=comment)
            geneCommentary.gene = gene
            geneCommentary.gene_commentary = gene_commentary
            self.session.add(geneCommentary)
            self.session.flush()
        return geneCommentary
    
    def getGeneCommentaryType(self, gene_commentary_type_id=None,
        commentary_type=None):
        """
        2012.5.15
            moved from transfac/src/GeneASNXML2gene_mapping.py
        """
        if gene_commentary_type_id:
            gene_commentary_type = self.checkIfEntryInTable(
                TableClass=GeneCommentaryType, entry_id=gene_commentary_type_id)
        elif commentary_type:
            gene_commentary_type = self.queryTable(GeneCommentaryType).\
                filter_by(type=commentary_type).first()
        else:
            gene_commentary_type = None
        if not gene_commentary_type:
            if self.debug:
                logging.warn("\t Gene-commentary_type %s not in db yet."%\
                    commentary_type)
            gene_commentary_type = GeneCommentaryType(type=commentary_type)
            if gene_commentary_type_id:
                gene_commentary_type.id = gene_commentary_type_id
            self.session.add(gene_commentary_type)
            self.session.flush()
        return gene_commentary_type
    
    def getGene(self, annot_assembly_id=None, locustag=None, tax_id=None, 
        chromosome=None, strand=None,\
        start=None, stop=None, type_of_gene=None, entrezgene_type_id=None, 
        chromosome_sequence_type_id=None,
        annot_assembly=None, ncbi_gene_id=None, gene_symbol=None, 
        synonyms=None, dbxrefs=None, map_location=None, description=None, 
        genomic_accession=None, genomic_version=None, genomic_gi=None, \
        symbol_from_nomenclature_authority=None,
        full_name_from_nomenclature_authority=None,\
        nomenclature_status=None, modification_date=None):
        """
        2012.4.26
using_table_options(UniqueConstraint('tax_id', 'locustag', 'chromosome', 
    'strand', 'start', 'stop', 'type_of_gene'))
        """
        if annot_assembly_id is None and annot_assembly:
            annot_assembly_id = annot_assembly.id
        if annot_assembly_id is None and chromosome and tax_id  and chromosome_sequence_type_id:
            annot_assembly = self.getAnnotAssembly(tax_id=tax_id, \
                sequence_type_id=chromosome_sequence_type_id, chromosome=chromosome)
            annot_assembly_id = annot_assembly.id
        
        if annot_assembly_id and start and stop and tax_id and chromosome and entrezgene_type_id:
            query = self.queryTable(Gene).filter_by(annot_assembly_id=annot_assembly.id).\
                filter_by(start=start).filter_by(stop=stop).\
                filter_by(tax_id=tax_id).filter_by(chromosome=chromosome).\
                filter_by(entrezgene_type_id=entrezgene_type_id)
            if type_of_gene:
                query = query.filter_by(type_of_gene=type_of_gene)
            gene = query.first()
        else:
            gene = None
        if gene is None:
            gene = Gene(annot_assembly_id=annot_assembly_id, tax_id=tax_id, 
                chromosome=chromosome, locustag=locustag, strand=strand,
                start=start, stop=stop, type_of_gene=type_of_gene, 
                entrezgene_type_id=entrezgene_type_id,\
                ncbi_gene_id=ncbi_gene_id, gene_symbol=gene_symbol, 
                synonyms=synonyms, dbxrefs=dbxrefs, map_location=map_location,
                description=description, genomic_accession=genomic_accession, 
                genomic_version=genomic_version, genomic_gi=genomic_gi, \
                symbol_from_nomenclature_authority=symbol_from_nomenclature_authority,
                full_name_from_nomenclature_authority=full_name_from_nomenclature_authority,
                nomenclature_status=nomenclature_status,
                modification_date=modification_date)
            gene.annot_assembly = annot_assembly
            self.session.add(gene)
            self.session.flush()
        return gene
    
    def constructGeneSegments(self, box_ls=[], gene_commentary=None, gi=None,
        commentary_type=None):
        """
        2012.5.15
            add argument commentary_type to stop
                replicating gene_commentary.gene_commentary_type
        2012.4.26
        """
        gene_segment_ls = []
        if commentary_type:
            gene_commentary_type = self.getGeneCommentaryType(
                commentary_type=commentary_type)
        else:
            gene_commentary_type = gene_commentary.gene_commentary_type
        for box in box_ls:
            start, stop = box[:2]
            
            gene_segment = GeneSegment(start=start, stop=stop, gi=gi,
                gene_commentary_type=gene_commentary_type)
            gene_segment.gene_commentary = gene_commentary
            self.session.add(gene_segment)
            gene_segment_ls.append(gene_segment)
        self.session.flush()
        return gene_segment_ls
    
    def run(self):
        """
        """
        if self.debug:
            import pdb
            pdb.set_trace()
        
        if self.geneAnnotationPickleFname and self.tax_id:
            import pickle
            pickle_fname = os.path.expanduser(self.geneAnnotationPickleFname)
            #2012.3.26 stopped using '~/at_gene_model_pickelf',
            if os.path.isfile(pickle_fname):
                #2011-1-20 check if the pickled file already exists or not
                logging.error("File %s already exists, no gene model pickle output."\
                    %(pickle_fname))
                sys.exit(2)
            else:
                #2008-10-01	get gene model and pickle it into a file
                gene_id2model, chr_id2gene_id_ls, geneSpanRBDict = \
                    self.get_gene_id2model(tax_id=self.tax_id)
                gene_annotation = PassingData()
                gene_annotation.gene_id2model = gene_id2model
                gene_annotation.chr_id2gene_id_ls = chr_id2gene_id_ls
                gene_annotation.geneSpanRBDict = geneSpanRBDict
                picklef = open(os.path.expanduser(pickle_fname), 'w')
                pickle.dump(gene_annotation, picklef, -1)
                picklef.close()

def get_entrezgene_annotated_anchor(curs, tax_id,
    entrezgene_mapping_table='genome.gene',\
    annot_assembly_table='genome.annot_assembly'):
    """
    2011-1-25
        moved from annot.bin.common
        change e.gene_id to e.id
    2010-8-1 make sure chromosome is not null
    
    2008-08-12
        deal with genes with no strand info
        only start of the gene gets into chromosome2anchor_gene_tuple_ls. not stop.
    12-11-05
    
    12-17-05
        for TFBindingSiteParse.py
    """
    sys.stderr.write("Getting entrezgene_annotated_anchor ...")
    chromosome2anchor_gene_tuple_ls = {}
    gene_id2coord = {}
    curs.execute(f"select e.genomic_gi, a.chromosome, e.id, e.strand, "
        f"e.start, e.stop from {entrezgene_mapping_table} e, "
        f"{annot_assembly_table} a where e.genomic_gi = a.gi and "
        f"e.tax_id={tax_id} and a.chromosome is not null")
        #2010-8-1 make sure chromosome is not null
    
    #curs.execute("fetch 5000 from eaa_crs")
    rows = curs.fetchall()
    counter = 0#2010-8-1
    #while rows:
    for row in rows:
        genomic_gi, chromosome, gene_id, strand, start, stop = row
        gene_id = int(gene_id)
        if strand=='1' or strand=='+1' or strand=='+':
            strand = '+'
        elif strand=='-1' or strand=='-':
            strand = '-'
        else:
            strand = strand
        
        if chromosome not in chromosome2anchor_gene_tuple_ls:
            chromosome2anchor_gene_tuple_ls[chromosome] = []
        
        chromosome2anchor_gene_tuple_ls[chromosome].append((start, gene_id))
        gene_id2coord[gene_id] = (start, stop, strand, genomic_gi)
        #curs.execute("fetch 5000 from eaa_crs")
        #rows = curs.fetchall()
        counter += 1
    for chromosome in chromosome2anchor_gene_tuple_ls:	#sort the list
        chromosome2anchor_gene_tuple_ls[chromosome].sort()
    #curs.execute("close eaa_crs")
    sys.stderr.write("%s genes (including AS-isoforms). Done.\n"%counter)
    return chromosome2anchor_gene_tuple_ls, gene_id2coord

if __name__ == '__main__':
    import sys, os
    from palos import ProcessOptions
    main_class = GenomeDatabase
    po = ProcessOptions(sys.argv, main_class.option_default_dict, \
        error_doc=main_class.__doc__)
    
    instance = main_class(**po.long_option2value)
    instance.setup()
    instance.run()
    """
    if instance.debug:
        import pdb
        pdb.set_trace()
    
    import pickle
    #2011-1-20 check if the pickled file already exists or not
    pickle_fname = po.arguments[0]
    #2012.3.26 stopped using '~/at_gene_model_pickelf', instead use the 1st argument
    if os.path.isfile(os.path.expanduser(pickle_fname)):
        logging.error(f"File {pickle_fname} already exists, "
            "no gene model pickle output."
    else:
        #2008-10-01	get gene model and pickle it into a file
        gene_id2model, chr_id2gene_id_ls, geneSpanRBDict = \
            instance.get_gene_id2model(tax_id=3702)
        gene_annotation = PassingData()
        gene_annotation.gene_id2model = gene_id2model
        gene_annotation.chr_id2gene_id_ls = chr_id2gene_id_ls
        gene_annotation.geneSpanRBDict = geneSpanRBDict
        picklef = open(os.path.expanduser(pickle_fname), 'w')
        pickle.dump(gene_annotation, picklef, -1)
        picklef.close()
    """
    import sqlalchemy as sql
    #print dir(Gene)
    #print Gene.table.c.keys()
    #results = instance.session.query(Gene)
    #results = instance.session.execute(sql.select([Gene.table]))
    #print dir(results)
    #print dir(Gene.query)
    
    #import pdb
    #pdb.set_trace()
    i = 0
    block_size = 10
    rows = Gene.query.offset(i).limit(block_size)
    print(dir(rows))
    while rows.count()!=0:
        print(rows.count())
        for row in rows:
            i += 1
            print(row.id, row.ncbi_gene_id, row.gene_symbol)
        if i>=5*block_size:
            break
        rows = Gene.query.offset(i).limit(block_size)


#!/usr/bin/env python3
"""
Coordinates in this module are assumed to be 1-based (and inclusive)
    i.e. start=1,stop=100 means [1,100].
2009-10-31
    module for CNV (copy-number-variation)-related functions & classes
"""

import os, sys
import fileinput
import numpy
from palos.polymorphism.SNP import GenomeWideResult, DataObject
from palos.utils import getColName2IndexFromHeader, figureOutDelimiter, PassingData

def get_overlap_ratio(span1_ls=None, span2_ls=None, isDataDiscrete=True):
    """
    isDataDiscrete: boolean
        True, numbers in span are discrete, like 1,2,3 (i.e. chromosomal position). then length of [1,2] is 2 because each number
            occupies 1 on axis.
        False, numbers in span are continuous, then length of [1,2] is 1 because each number occupies infinitesimal amount on axis.
        This explains why UCSC genome browser adopts a [1,3) notation, instead of a [1,2] notation.
    2013.1.27 add total_span, overlapFraction, and return PassingData (data structure)
    2012.5.17 bugfix. overlap_length can't be negative (hard-set to zero).
    2010-8-18
        figure out the overlap coordinates
    2010-8-2
        add +1 in "max(0, segment_stop_pos - qc_start+1)" to recognize even 1-bp overlapping
        bug fix: span1_ls and/or span2_ls have same value on start & stop.
    2010-7-30
        doc: if two spans have no overlap, it returns 0,0
    2010-2-11
        swap overlapFraction1 and overlapFraction2 so that they match the span1_ls and span2_ls
    2009-12-13
        calculate the two overlap ratios for two segments
    """
    segment_start_pos, segment_stop_pos = span1_ls
    qc_start, qc_stop = span2_ls
    #2010-8-18 figure out the overlap coordinates
    overlap_start_pos = max(qc_start, segment_start_pos)
    total_start_pos = min(qc_start, segment_start_pos)
    
    overlap_stop_pos = min(qc_stop, segment_stop_pos)
    total_stop_pos = max(qc_stop, segment_stop_pos)
    
    
    total_span = abs(total_stop_pos-total_start_pos)
    if isDataDiscrete:
        total_start_pos += 1	#2013.1.27 add +1 for discrete data
    
    if overlap_start_pos>overlap_stop_pos:
        overlap_start_pos = None
        overlap_stop_pos = None
    if isDataDiscrete:
        # accomodates 6 scenarios
        overlap_length = max(0, segment_stop_pos - qc_start+1) - \
            max(0, segment_stop_pos - qc_stop) - max(0, segment_start_pos - qc_start)
    else:
        overlap_length = max(0, segment_stop_pos - qc_start) - \
            max(0, segment_stop_pos - qc_stop) - max(0, segment_start_pos - qc_start)
    overlap_length = float(overlap_length)
    if segment_stop_pos == segment_start_pos:	#2010-8-2 bugfix
        if overlap_length>0:
            overlapFraction1 = 1.0
        else:
            overlap_length = 0	#2012.5.17 bugfix. can't be negative
            overlapFraction1 = 0.0
    else:
        if isDataDiscrete:
            segmentLength = segment_stop_pos-segment_start_pos+1    # add +1?
        else:
            segmentLength = segment_stop_pos-segment_start_pos
        overlapFraction1 = overlap_length/segmentLength
    if qc_stop == qc_start:
        if overlap_length>0:
            overlapFraction2 = 1.0
        else:
            overlap_length = 0  #can't be negative
            overlapFraction2 = 0.0
    else:
        if isDataDiscrete:
            qcLength = qc_stop-qc_start+1	#2013.1.27 add +1 ?
        else:
            qcLength = qc_stop-qc_start
        overlapFraction2 = overlap_length/qcLength
    overlapFraction = float(overlap_length)/total_span	#2013.1.27
    return PassingData(overlapFraction1=overlapFraction1, overlapFraction2=overlapFraction2, \
            overlap_length=overlap_length, \
            overlapFraction=overlapFraction,\
            overlap_start_pos=overlap_start_pos, overlap_stop_pos=overlap_stop_pos, \
            total_start_pos=total_start_pos, total_stop_pos=total_stop_pos,\
            total_span = total_span)

def is_reciprocal_overlap(span1_ls=None, span2_ls=None, min_reciprocal_overlap=0.6, \
                          isDataDiscrete=True):
    """
    2009-12-12
        return True if both overlap ratios are above the min_reciprocal_overlap
    """
    overlapData = get_overlap_ratio(span1_ls=span1_ls, span2_ls=span2_ls, \
                                    isDataDiscrete=isDataDiscrete)
    overlapFraction1 = overlapData.overlapFraction1
    overlapFraction2 = overlapData.overlapFraction2
    if overlapFraction1>=min_reciprocal_overlap and overlapFraction2>=min_reciprocal_overlap:
        return True
    else:
        return False

class CNVSegmentBinarySearchTreeKey(object):
    """
    A key designed to represent a CNV segment in the node of a Red-Black tree (algorithm/RBTree.py)
        or binary search tree (BinarySearchTree.py, less efficient than RBTree.py).

    It has custom comparison function based on the is_reciprocal_overlap() function.
    
    2013.06.25 added argument isDataDiscrete to treat discrete and continuous coordinates differently
    2010-8-2
        Added argument keywords to __init__()
    2009-12-12
        Init.
    """
    def __init__(self, chromosome=None, span_ls=None, min_reciprocal_overlap=0.6, \
            isDataDiscrete=True, **keywords):
        self.chromosome = chromosome
        self.span_ls = span_ls
        self.min_reciprocal_overlap = min_reciprocal_overlap
        self.isDataDiscrete= isDataDiscrete
        
        self.start = self.span_ls[0]
        if len(self.span_ls)>1:
            self.stop = self.span_ls[1]
            if self.start is not None and self.stop is not None:
                self.span = abs(self.stop-self.start)
            else:
                self.span = 0
        else:
            self.stop = self.start
            self.span = 0
        if self.stop<self.start:
            sys.stderr.write("Warning: not supposed to happen. stop %s is smaller than start %s.\n"%(self.stop, self.start))
        for key, value in keywords.items():
            setattr(self, key, value)
    
    def __eq__(self, other):
        """
        equal: self == other
        2010-2-11
            fix a bug when len(self.span_ls)>1 and len(other.span_ls)==1.
            it was completely wrong before.
        2009-12-12
        """
        if self.chromosome==other.chromosome:
            if len(self.span_ls)==1:
                if len(other.span_ls)==1:
                    return self.span_ls[0]==other.span_ls[0]
                elif len(other.span_ls)>1:
                    return self.span_ls[0]>=other.span_ls[0] and self.span_ls[0]<=other.span_ls[1]	# equal if self is within the "other" segment
                else:
                    return None
            elif len(self.span_ls)>1:
                if len(other.span_ls)==1:	# self is a segment. other is a point position.
                    return self.span_ls[0]<=other.span_ls[0] and self.span_ls[1]>=other.span_ls[0]	# if self includes the point position, yes it's equal
                elif len(other.span_ls)>1:
                    # need to calculate min_reciprocal_overlap
                    return is_reciprocal_overlap(self.span_ls, other.span_ls, \
                                min_reciprocal_overlap=self.min_reciprocal_overlap,\
                                isDataDiscrete=self.isDataDiscrete)
                    #return self.span_ls[1]<other.span_ls[0]	# whether the stop of this segment is ahead of the start of other
                else:
                    return None
        else:
            return False
    
    def __ge__(self, other):
        """
        2009-12-12
        greater than or equal: self >= other.
        """
        if self.chromosome==other.chromosome and len(self.span_ls)>1 and len(other.span_ls)>1:
            if is_reciprocal_overlap(self.span_ls, other.span_ls, \
                    min_reciprocal_overlap=self.min_reciprocal_overlap,\
                    isDataDiscrete=self.isDataDiscrete):
                return True
            else:
                return not self.__lt__(other)
        else:
            return not self.__lt__(other)
        
    def __gt__(self, other):
        """
        2009-12-12
        greater than: self > other.
        """
        if self.chromosome==other.chromosome and len(self.span_ls)>1 and len(other.span_ls)>1:
            if is_reciprocal_overlap(self.span_ls, other.span_ls, \
                    min_reciprocal_overlap=self.min_reciprocal_overlap,\
                    isDataDiscrete=self.isDataDiscrete):
                ## too much overlap. The two objects are equal (not self > other). 
                return False
            else:
                return not self.__le__(other)
        else:
            return not self.__le__(other)

    def __hash__(self):
        """
        make it hashable
        """
        return hash((self.chromosome, tuple(self.span_ls)))
    

    def __lt__(self, other):
        """
        less than: self < other.
        2009-12-12
        """
        if self.chromosome==other.chromosome:
            if len(self.span_ls)==1:
                if len(other.span_ls)==1:
                    return self.span_ls[0]<other.span_ls[0]
                elif len(other.span_ls)>1:
                    return self.span_ls[0]<other.span_ls[0]
                else:
                    return None
            elif len(self.span_ls)>1:
                if len(other.span_ls)==1:
                    return self.span_ls[0]<other.span_ls[0]
                elif len(other.span_ls)>1:
                    if is_reciprocal_overlap(self.span_ls, other.span_ls, \
                        min_reciprocal_overlap=self.min_reciprocal_overlap,\
                        isDataDiscrete=self.isDataDiscrete):
                        ## too much overlap. The two objects are equal (not self < other). 
                        return False
                    #if overlapFraction1>0 and overlapFraction1<=1 and  overlapFraction2>0 and overlapFraction2<=1:	# there's overlap between two segments 
                    #	return False
                    else:
                        # whether the stop of this segment is ahead of the start of other
                        return self.span_ls[0]<other.span_ls[0]
                else:
                    return None
        else:
            return self.chromosome<other.chromosome
    
    def __le__(self, other):
        """
        less than or equal: self <= other.
        2009-12-12
        """
        if self.chromosome==other.chromosome:
            if len(self.span_ls)==1:	# self is a point
                if len(other.span_ls)==1:
                    return self.span_ls[0]<=other.span_ls[0]
                elif len(other.span_ls)>1:
                    return self.span_ls[0]<=other.span_ls[0]
                else:
                    return None
            elif len(self.span_ls)>1:
                if len(other.span_ls)==1:
                    return self.span_ls[0]<=other.span_ls[0]
                elif len(other.span_ls)>1:
                    if is_reciprocal_overlap(self.span_ls, other.span_ls, \
                            min_reciprocal_overlap=self.min_reciprocal_overlap,\
                            isDataDiscrete=self.isDataDiscrete):
                        return True
                    else:
                        # whether the stop of this segment is ahead of the start of other
                        return self.span_ls[0]<=other.span_ls[0]
                else:
                    return None
        else:
            return self.chromosome < other.chromosome
    
    def __ne__(self, other):
        """
        not equal: self != other.
        2009-12-12
        """
        return not self.__eq__(other)
    
    def __str__(self):
        """
        2009-12-13
        """
        return "chromosome: %s, span_ls: %s"%(self.chromosome, repr(self.span_ls))
    
    def getKey(self):
        """
        2012.11.20 return a tuple that is dictionary-able.
        """
        return (self.chromosome, self.start, self.stop)


class SegmentTreeNodeKey(CNVSegmentBinarySearchTreeKey):
    """
    2013.06.25 added argument isDataDiscrete to treat discrete and continuous coordinates differently
    2010-1-28
        tree node key which counts both is_reciprocal_overlap()=True and "other" being wholly embedded in "self" as equal
        
        similar purpose as the function leftWithinRightAlsoEqualCmp().
        
        One strange thing (???) about who is "self", who is "other" in the __eq__() below.
            In the situation you have a new CNVSegmentBinarySearchTreeKey key (call it "segmentKey"),
            and want to test if this "segmentKey" is in the tree or not as in "if segmentKey in tree: ...".
             
            If the tree is comprised of CNVSegmentBinarySearchTreeKey instances, self=segmentKey, other=nodes in the tree.
            If the tree is comprised of SegmentTreeNodeKey instances, self=nodes in the tree, other=segmentKey.
                That's why here used the condition that "other"'s overlap ratio (overlapFraction2) ==1..
            
        
    """
    def __init__(self, chromosome=None, span_ls=None, min_reciprocal_overlap=0.6, isDataDiscrete=True):
        CNVSegmentBinarySearchTreeKey.__init__(self, chromosome=chromosome, span_ls=span_ls, \
                                            min_reciprocal_overlap=min_reciprocal_overlap, isDataDiscrete=isDataDiscrete)
    
    def __eq__(self, other):
        """
        2010-1-28
        """
        return rightWithinLeftAlsoEqual(self, other)


class CNVCompareBySmallOverlapRatio(object):
    """
    A class to compare two CNVSegmentBinarySearchTreeKey based on min_reciprocal_overlap
    2010-8-16
    """
    def __init__(self, min_reciprocal_overlap = 0.4, isDataDiscrete=True):
        """
        2010-8-16
        """
        self.min_reciprocal_overlap = min_reciprocal_overlap
        self.isDataDiscrete = isDataDiscrete
    
    def cmp(self, key1, key2):
        if self.eq(key1, key2, min_reciprocal_overlap=self.min_reciprocal_overlap):
            return 0
        elif self.lt(key1, key2, min_reciprocal_overlap=self.min_reciprocal_overlap):
            return -1
        else:
            return +1
    
    @classmethod
    def eq(cls, key1, key2, min_reciprocal_overlap=0.4,isDataDiscrete=True):
        if key1.chromosome==key2.chromosome:
            if len(key1.span_ls)==1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]==key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    return key1.span_ls[0]>=key2.span_ls[0] and key1.span_ls[0]<=key2.span_ls[1]	# equal if self is within the "other" segment
                else:
                    return None
            elif len(key1.span_ls)>1:
                if len(key2.span_ls)==1:
                    # self is a segment. other is a point position.
                    return key1.span_ls[0]<=key2.span_ls[0] and key1.span_ls[1]>=key2.span_ls[0]	# if self includes the point position, yes it's equal
                elif len(key2.span_ls)>1:
                    # need to calculate min_reciprocal_overlap
                    return is_reciprocal_overlap(key1.span_ls, key2.span_ls, \
                                min_reciprocal_overlap=min_reciprocal_overlap,\
                                isDataDiscrete=isDataDiscrete)
                else:
                    return None		#2010-8-16 shouldn't be here
        else:
            return False
        
        return True
    
    @classmethod
    def lt(cls, key1, key2, min_reciprocal_overlap=0.4, isDataDiscrete=True):
        if key1.chromosome==key2.chromosome:
            if len(key1.span_ls)==1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]<key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    return key1.span_ls[0]<key2.span_ls[0]
                else:
                    return None
            elif len(key1.span_ls)>1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]<key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    if is_reciprocal_overlap(key1.span_ls, key2.span_ls, \
                                min_reciprocal_overlap=min_reciprocal_overlap,\
                                isDataDiscrete=isDataDiscrete): 
                        return False
                    else:
                        # whether the start of key1 is ahead of the start of key2
                        return key1.span_ls[0]<key2.span_ls[0]
                else:
                    return None	#shouldn't be here
        else:
            return key1.chromosome<key2.chromosome
    
    @classmethod
    def gt(cls, key1, key2, min_reciprocal_overlap=0.4):
        """
        2010-8-16
            not used
        """
        pass

#2011-3-24 make an alias
CNVCompare = CNVCompareBySmallOverlapRatio


class CNVCompareByOverlapLen(object):
    """
    2011-3-16
    A class to compare two CNVSegmentBinarySearchTreeKey according to any arbitrary min_overlap_len.
    """
    def __init__(self, min_overlap_len = 1):
        """
        2011-3-16
        """
        self.min_overlap_len = min_overlap_len
    
    def cmp(self, key1, key2):
        if self.eq(key1, key2, min_overlap_len=self.min_overlap_len):
            return 0
        elif self.lt(key1, key2, min_overlap_len=self.min_overlap_len):
            return -1
        else:
            return +1
    
    @classmethod
    def eq(cls, key1, key2, min_overlap_len=1, isDataDiscrete=True):
        if key1.chromosome==key2.chromosome:
            if len(key1.span_ls)==1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]==key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    return key1.span_ls[0]>=key2.span_ls[0] and key1.span_ls[0]<=key2.span_ls[1]	# equal if self is within the "other" segment
                else:
                    return None
            elif len(key1.span_ls)>1:
                if len(key2.span_ls)==1:	# self is a segment. other is a point position.
                    return key1.span_ls[0]<=key2.span_ls[0] and key1.span_ls[1]>=key2.span_ls[0]	# if self includes the point position, yes it's equal
                elif len(key2.span_ls)>1:
                    # need to calculate min_reciprocal_overlap
                    overlapData = get_overlap_ratio(key1.span_ls, key2.span_ls, \
                                                    isDataDiscrete=isDataDiscrete)
                    overlapFraction1 = overlapData.overlapFraction1
                    overlapFraction2 = overlapData.overlapFraction2
                    overlap_length = overlapData.overlap_length
                    if overlap_length>=min_overlap_len:	#should be equal,
                        return True
                    else:
                        return False
                else:
                    return None		#2010-8-16 shouldn't be here
        else:
            return False
        
        return True
    
    @classmethod
    def lt(cls, key1, key2, min_overlap_len=0.4, isDataDiscrete=True):
        """
        2011-3-16
        """
        if key1.chromosome==key2.chromosome:
            if len(key1.span_ls)==1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]<key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    return key1.span_ls[0]<key2.span_ls[0]
                else:
                    return None
            elif len(key1.span_ls)>1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]<key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    overlapData = get_overlap_ratio(key1.span_ls, key2.span_ls, \
                                                    isDataDiscrete=isDataDiscrete)
                    overlapFraction1 = overlapData.overlapFraction1
                    overlapFraction2 = overlapData.overlapFraction2
                    overlap_length = overlapData.overlap_length
                    if overlap_length>=min_overlap_len:	#should be equal,
                        return False
                    else:
                        # whether the start of key1 is ahead of the start of key2
                        return key1.span_ls[0]<key2.span_ls[0]
                else:
                    return None	#shouldn't be here
        else:
            return key1.chromosome<key2.chromosome
    
    @classmethod
    def gt(cls, key1, key2, min_reciprocal_overlap=0.4):
        """
        2011-3-16
            not used
        """
        pass

class CNVCompareByBigOverlapRatio(object):
    """
    2011-3-16
    Similar to CNVCompareBySmallOverlapRatio, but use max(overlapFraction1, overlapFraction2) \
        to compare against min_overlap_len, \
        rather than min(overlapFraction1, overlapFraction2)
    """
    def __init__(self, min_reciprocal_overlap = 0.4):
        """
        2010-8-16
        """
        self.min_reciprocal_overlap = min_reciprocal_overlap
    
    def cmp(self, key1, key2):
        if self.eq(key1, key2, min_reciprocal_overlap=self.min_reciprocal_overlap):
            return 0
        elif self.lt(key1, key2, min_reciprocal_overlap=self.min_reciprocal_overlap):
            return -1
        else:
            return +1
    
    @classmethod
    def eq(cls, key1, key2, min_reciprocal_overlap=0.4, isDataDiscrete=True):
        if key1.chromosome==key2.chromosome:
            if len(key1.span_ls)==1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]==key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    return key1.span_ls[0]>=key2.span_ls[0] and key1.span_ls[0]<=key2.span_ls[1]	# equal if self is within the "other" segment
                else:
                    return None
            elif len(key1.span_ls)>1:
                if len(key2.span_ls)==1:	# self is a segment. other is a point position.
                    return key1.span_ls[0]<=key2.span_ls[0] and key1.span_ls[1]>=key2.span_ls[0]	# if self includes the point position, yes it's equal
                elif len(key2.span_ls)>1:
                    overlapData = get_overlap_ratio(key1.span_ls, key2.span_ls, isDataDiscrete=isDataDiscrete)
                    overlapFraction1 = overlapData.overlapFraction1
                    overlapFraction2 = overlapData.overlapFraction2
                    overlap_length = overlapData.overlap_length
                    recip_overlap = max(overlapFraction1, overlapFraction2)
                    if recip_overlap>=min_reciprocal_overlap:	#should be equal,
                        return True
                    else:
                        return False
                else:
                    return None		#2010-8-16 shouldn't be here
        else:
            return False
        
        return True
    
    @classmethod
    def lt(cls, key1, key2, min_reciprocal_overlap=0.4, isDataDiscrete=True):
        if key1.chromosome==key2.chromosome:
            if len(key1.span_ls)==1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]<key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    return key1.span_ls[0]<key2.span_ls[0]
                else:
                    return None
            elif len(key1.span_ls)>1:
                if len(key2.span_ls)==1:
                    return key1.span_ls[0]<key2.span_ls[0]
                elif len(key2.span_ls)>1:
                    overlapData = get_overlap_ratio(key1.span_ls, key2.span_ls)
                    overlapFraction1 = overlapData.overlapFraction1
                    overlapFraction2 = overlapData.overlapFraction2
                    recip_overlap = max(overlapFraction1, overlapFraction2, isDataDiscrete=isDataDiscrete)
                    if recip_overlap>=min_reciprocal_overlap:	#should be equal,
                        return False
                    else:
                        return key1.span_ls[0]<key2.span_ls[0]	# whether the start of key1 is ahead of the start of key2
                else:
                    return None	#shouldn't be here
        else:
            return key1.chromosome<key2.chromosome
    
    @classmethod
    def gt(cls, key1, key2, min_reciprocal_overlap=0.4):
        """
        2010-8-16
            not used
        """
        pass

def leftWithinRightAlsoEqual(key1, key2, isDataDiscrete=True):
    """
    2010-1-28
    Besides key1.__eq__(key2), if key1 is embedded in key2, it's also regarded as equal.
    This function is solely used in leftWithinRightAlsoEqualCmp().
    """
    equalResult = key1.__eq__(key2)
    if equalResult:
        return equalResult
    else:
        if len(key1.span_ls)==2 and len(key2.span_ls)==2:
            overlapData = get_overlap_ratio(key1.span_ls, key2.span_ls, isDataDiscrete=isDataDiscrete)
            overlapFraction1 = overlapData.overlapFraction1
            overlapFraction2 = overlapData.overlapFraction2
            if overlapFraction1==1.:
                return True
            else:
                return equalResult
        else:
            return equalResult

def rightWithinLeftAlsoEqual(key1, key2, isDataDiscrete=True):
    """
    2010-1-28
        Besides CNVSegmentBinarySearchTreeKey.__eq__(), if key2 is embedded in key1, it's also regarded as equal.
        this function is solely used in leftWithinRightAlsoEqualCmp().
    """
    equalResult = key1.__eq__(key2)
    if equalResult:
        return equalResult
    else:
        if len(key1.span_ls)==2 and len(key2.span_ls)==2:
            overlapData = get_overlap_ratio(key1.span_ls, key2.span_ls, isDataDiscrete=isDataDiscrete)
            overlapFraction1 = overlapData.overlapFraction1
            overlapFraction2 = overlapData.overlapFraction2
            overlap_length = overlapData.overlap_length
            if overlapFraction2==1.:	# 2010-1-28 added the overlapFraction2==1.
                return True
            else:
                return equalResult
        else:
            return equalResult

def leftWithinRightAlsoEqualCmp(key1, key2):
    """
    2010-1-28
    A cmp function for RBDict to compare CNVSegmentBinarySearchTreeKey or SegmentTreeNodeKey.
    key1 is the new entry or query. key2 is the one that is already in RBDict. 
    
    If key1 is embedded in key2, it's also regarded as equal.
    
    similar purpose as the class SegmentTreeNodeKey()

    Example:
        Add this to RBDict as in "rb_dict = RBDict(cmpfn=leftWithinRightAlsoEqualCmp)".
    """
    if leftWithinRightAlsoEqual(key1, key2):
        return 0
    elif key1 > key2:
        return 1
    else: #key1 < key2
        return -1

def rightWithinLeftAlsoEqualCmp(key1, key2):
    """
    2010-1-28
    A cmp function for RBDict to compare CNVSegmentBinarySearchTreeKey or SegmentTreeNodeKey.
    key1 is the new entry or query. key2 is the one that is already in RBDict.
    If key2 is embedded in key1, it's also regarded as equal.
    
    similar purpose as the class SegmentTreeNodeKey().
    
    Use this with RBDict: "rb_dict = RBDict(cmpfn=rightWithinLeftAlsoEqualCmp)".
    
    
    """
    if rightWithinLeftAlsoEqual(key1, key2):
        return 0
    elif key1 > key2:
        return 1
    else:  #key1 < key2
        return -1

def getCNVDataFromFileInGWA(input_fname_ls, array_id, max_amp=-0.33, min_amp=-0.33, \
        min_size=50, min_no_of_probes=None, \
        report=False, chr=None, start=None, stop=None, filter_gada_by_cutoff=True):
    """
    2010-6-5
        data_obj.genome_wide_result_name = gwr_name
        add ecotyp_id into data_obj.comment
    2010-6-3
        add argument chr, start, stop to restrict fragments within those
    2009-10-31
        get deletion (below max_amp) or duplication (above min_amp) from files (output by RunGADA.py)
    """
    sys.stderr.write("Getting CNV calls for array %s, min_size %s, min_no_of_probes %s from %s ..."%\
        (array_id, min_size, min_no_of_probes, repr(input_fname_ls)))
    
    gwr_name = "(a-id %s)"%(array_id)
    gwr = GenomeWideResult(name=gwr_name)
    gwr.data_obj_ls = []	#list and dictionary are crazy references.
    gwr.data_obj_id2index = {}
    genome_wide_result_id = id(gwr)
    
    amp_ls = []
    array_id2array = {}
    counter = 0
    real_counter = 0
    no_of_segments = 0
    input_handler = fileinput.input(input_fname_ls)
    header = input_handler.readline().strip().split('\t')
    col_name2index = getColName2IndexFromHeader(header)
    ecotype_id = None
    for line in input_handler:
        if line.find("array_id")!=-1:
            continue
        line = line.strip()
        row = line.split('\t')
        cnv_array_id = int(row[col_name2index['array_id']])
        cnv_ecotype_id = int(row[col_name2index.get('ecotype_id', col_name2index['array_id'])])
        counter += 1
        if cnv_array_id==array_id:
            no_of_segments += 1
            if ecotype_id is None:
                ecotype_id = cnv_ecotype_id
            start_probe = row[col_name2index['start_probe']].split('_')	# split chr_pos
            start_probe = map(int, start_probe)
            start_probe_id = row[col_name2index.get('start_probe_id', col_name2index['start_probe'])]
            
            stop_probe = row[col_name2index['end_probe']].split('_')
            stop_probe = map(int, stop_probe)
            if start_probe[0]!=stop_probe[0]:	#fragments on different chromosome. ignore.
                continue
            end_probe_id = row[col_name2index.get('end_probe_id', col_name2index['end_probe'])]
            
            no_of_probes = int(row[col_name2index['length']])
            if min_no_of_probes is not None and no_of_probes<min_no_of_probes:
                continue
            amplitude = float(row[col_name2index['amplitude']])
            segment_chromosome = start_probe[0]
            segment_start_pos = start_probe[1]-12
            segment_stop_pos = stop_probe[1]+12
            segment_length = abs(segment_stop_pos-segment_start_pos)
            if chr is not None and segment_chromosome!=chr:
                continue
            if start is not None and (segment_start_pos<start or segment_stop_pos<start):
                continue
            if stop is not None and (segment_start_pos>stop or segment_stop_pos>stop):
                continue
            if min_size is not None and segment_length<min_size:
                continue
            if amplitude<=max_amp:
                color = 'r'
            elif amplitude>=min_amp:
                color = 'b'
            else:
                color = 'g'
            if amplitude<=max_amp or amplitude>=min_amp or filter_gada_by_cutoff == False:
                real_counter += 1
                data_obj = DataObject(chromosome=segment_chromosome,
                    position=segment_start_pos, stop_position=segment_stop_pos, \
                    value=amplitude, color=color)
                data_obj.comment = 'ecotype-id %s, start probe-id %s, end probe-id %s, no of probes %s, size=%s'%\
                    (ecotype_id, start_probe_id, end_probe_id, no_of_probes, segment_length)
                data_obj.genome_wide_result_name = gwr_name
                data_obj.genome_wide_result_id = genome_wide_result_id
                gwr.add_one_data_obj(data_obj)
                
        if report and counter%10000==0:
            sys.stderr.write('%s%s\t%s\t%s'%('\x08'*80, counter, no_of_segments, real_counter))
    sys.stderr.write("\n")
    
    if gwr.max_value<3:	# insertion at y=3
        gwr.max_value=3
    if gwr.min_value>-1:	# deletion at y = -1
        gwr.min_value = -1
    gwr.name = '%s '%ecotype_id +  gwr.name
    setattr(gwr, 'ecotype_id', ecotype_id)
    sys.stderr.write(" %s segments. Done.\n"%(len(gwr.data_obj_ls)))
    return gwr

def mergeTwoSegments(segment, next_segment):
    """
    2010-7-19
    called by getCNVQCDataFromDB()
    
    segment = (row.chromosome, row.start, row.stop, row.size_affected, \
        row.no_of_probes_covered, row.copy_number, row.id, row.ecotype_id)
    
    copied from CNV.mergeTwoCNVQCSegments() in variation/src/misc.py
    """
    new_segment_stop = max(segment[2], next_segment[2])
    
    # to get the no_of_probes_covered for the merged segment
    if segment[4] is not None and next_segment[4] is not None:
        overlapping_len = float(segment[2]-next_segment[1])
        # it could be negative, which would end up increasing the no_of_probes_covered
        import numpy
        # estimate the number of probes in the overlapping region for each segment, take the average
        no_of_probes_in_the_overlapping_of_s1 = int(round(segment[4]*overlapping_len/segment[3]))
        no_of_probes_in_the_overlapping_of_s2 = int(round(next_segment[4]*overlapping_len/next_segment[3]))
        no_of_probes_in_the_overlapping = numpy.mean(
            [no_of_probes_in_the_overlapping_of_s1, no_of_probes_in_the_overlapping_of_s2])
        no_of_probes_covered = segment[4] + next_segment[4] - no_of_probes_in_the_overlapping
    else:
        no_of_probes_covered = None
    
    new_segment = [segment[0], segment[1], new_segment_stop,
        new_segment_stop-segment[1]+1, no_of_probes_covered] + \
        segment[5:]
    return new_segment

def mergeOverlappingAmongSegments(segment_ls):
    """
    2010-7-28
        called by turnSegmentGWRIntoRBDict()
    """
    segment_ls.sort()	#make sure in chromosmal order
    new_segment_ls = []
    for i in range(len(segment_ls)):
        segment = segment_ls[i]
        if len(new_segment_ls)>0:
            previous_segment = new_segment_ls[-1]
        else:
            previous_segment = None
        
        if previous_segment is not None and previous_segment[0]==segment[0] and previous_segment[2]>=segment[1]-1:
            previous_segment = mergeTwoSegments(previous_segment, segment)
            new_segment_ls[-1] = previous_segment
        else:
            new_segment_ls.append(segment)
    sys.stderr.write("%s segments merged into %s non-overlapping segments.\n"%(len(segment_ls), len(new_segment_ls)))
    return new_segment_ls

def turnSegmentGWRIntoRBDict(gwr, extend_dist=20000, min_reciprocal_overlap=0.6, report=True):
    """
    2010-7-28
        add functionality to merge overlapping segments.
        assuming data_obj_ls in gwr is in chromosomal order.
    2010-3-17
        extend_dist is used to enlarge the segments in each data_obj of gwr,
    """
    sys.stderr.write("Turning a segment-gwr (start-stop style) into an RBDict ...\n")

    segment_ls = []
    for data_obj in gwr.data_obj_ls:
        start = max(data_obj.position-extend_dist, 0)
        stop = data_obj.stop_position+extend_dist
        segment = [data_obj.chromosome, start, stop, stop-start+1, None]
        #the 4th bit (=None) is no_of_probes_covered.
        segment_ls.append(segment)
    
    segment_ls = mergeOverlappingAmongSegments(segment_ls)
    
    rbDict = RBDict(cmpfn=leftWithinRightAlsoEqualCmp)

    for segment in segment_ls:
        chromosome, start, stop = segment[:3]
        segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=chromosome,
            span_ls=[start, stop], \
            min_reciprocal_overlap=min_reciprocal_overlap)
        rbDict[segmentKey] = segment
    if report:
        print("\tDepth of rbDict: %d" % (rbDict.depth()))
        print("\tOptimum Depth: %f (%d) (%f%% depth efficiency)" % (rbDict.optimumdepth(), math.ceil(rbDict.optimumdepth()),
                math.ceil(rbDict.optimumdepth()) / rbDict.depth()))
    sys.stderr.write("%s objects converted.\n"%len(rbDict))
    return rbDict

def getProbeIntensityDataFromProbeXArray(input_fname, data_type=numpy.float32, \
        nonIntensityColumnLabelSet = set(['probes_id', 'chromosome', 'position']),\
        array_id=None):
    """
    2010-7-28
        bug-fix:
            make sure _array_id and array_id are of the same type.
    2010-6-28
        add argument array_id to get intensity only for that array.
    2010-5-23
        add argument nonIntensityColumnLabelSet to find out which columns harbor intensity data.
        Non-intensity columns (probes_id, chromosome, position) are not bound to any particular column.
    2010-3-18
        copied from CNVNormalize.get_input()
    2009-10-28
        switch the default data_type to numpy.float32 to save memory on 64bit machines
    2009-9-28
        add argument data_type to specify data type of data_matrix.
        default is numpy.float (numpy.float could be float32, float64, float128 depending on the architecture).
            numpy.double is also fine.
    2009-5-18
        become classmethod
    """
    sys.stderr.write("Getting Probe X Array intensity data from %s ..."%input_fname)
    import csv, subprocess
    reader = csv.reader(open(input_fname), delimiter=figureOutDelimiter(input_fname))
    commandline = 'wc -l %s'%input_fname
    command_handler = subprocess.Popen(commandline, shell=True,
        stderr=subprocess.PIPE, stdout=subprocess.PIPE,
        encoding='utf8')
    stdout_content, stderr_content = command_handler.communicate()
    if stderr_content:
        sys.stderr.write('stderr of %s: %s \n'%(commandline, stderr_content))
    no_of_rows = int(stdout_content.split()[0])-1
    
    header = next(reader)
    intensity_col_index_ls = []
    type_of_given_array_id = type(array_id)
    for i in range(len(header)):
        label = header[i]
        if label not in nonIntensityColumnLabelSet:
            # it's array_id if it's not in nonIntensityColumnLabelSet.
            _array_id = type_of_given_array_id(label)
            #2010-7-28 cast to whatever type array_id is of.
            if array_id is not None and _array_id!=array_id:
                #2010-6-28	skip this row
                continue
            intensity_col_index_ls.append(i)
    no_of_cols = len(intensity_col_index_ls)
    col_name2index = getColName2IndexFromHeader(header)
    data_matrix = numpy.zeros([no_of_rows, no_of_cols], data_type)
    probe_id_ls = []
    chr_pos_ls = []
    i=0
    for row in reader:
        probe_id = row[col_name2index["probes_id"]]
        probe_id_ls.append(probe_id)
        chromosome = row[col_name2index["chromosome"]]
        pos = row[col_name2index["position"]]
        chr_pos_ls.append((chromosome, pos))
        for j in intensity_col_index_ls:
            try:
                data_matrix[i][j-1] = float(row[j])
            except:
                data_matrix[i][j-1] = numpy.nan
        i += 1
    del reader
    sys.stderr.write("Done.\n")
    return data_matrix, probe_id_ls, chr_pos_ls, header

def getProbeIntensityDataFromArrayXProbe(input_fname, data_type=numpy.float32, \
    nonIntensityColumnLabelSet = None, array_id=None, no_of_arrays=None):
    """
    Similar to getProbeIntensityDataFromProbeXArray, but the input is in Probe X Array. 
    2010-7-28
        add argument no_of_arrays to skip the "wc -l" command
        bug-fix:
            make sure _array_id and array_id are of the same type.
    2010-6-28
        add argument array_id to get intensity only for that array.
    2010-5-25
    """
    sys.stderr.write("Getting Array X Probe intensity data from %s for array %s ...\n"%(input_fname, array_id))
    import csv, subprocess
    reader = csv.reader(open(input_fname), delimiter=figureOutDelimiter(input_fname))
    if array_id is not None:
        no_of_arrays = 1
    elif no_of_arrays is not None:
        no_of_arrays = no_of_arrays
    else:
        commandline = 'wc -l %s'%input_fname
        command_handler = subprocess.Popen(commandline, shell=True,
            stderr=subprocess.PIPE, stdout=subprocess.PIPE,
            encoding='utf8')
        stdout_content, stderr_content = command_handler.communicate()
        if stderr_content:
            sys.stderr.write('stderr of %s: %s \n'%(commandline, stderr_content))
        no_of_arrays = int(stdout_content.split()[0])-3
    
    header = next(reader)
    probe_id_ls = header[2:]
    chr_ls = next(reader)[2:]
    pos_ls = next(reader)[2:]
    chr_pos_ls = zip(chr_ls, pos_ls)
    
    no_of_probes = len(header)-2
    data_matrix = numpy.zeros([no_of_probes, no_of_arrays], data_type)
    array_id_ls = []
    i=0
    type_of_given_array_id = type(array_id)
    line_number = 0
    for row in reader:
        line_number += 1
        sys.stderr.write('%s%s'%("\x08"*80, line_number))
        
        if array_id is not None and len(array_id_ls)>0:
            #2010-8-6 already got the array, jump out.
            break
        
        _array_id = type_of_given_array_id(row[0])
        #2010-7-28 cast to whatever type array_id is of.
        if array_id is not None and _array_id!=array_id:
            #2010-6-28	skip this row
            continue
        ecotype_id = row[1]
        
        try:
            data_intensity_ls = map(float, row[2:])
            data_matrix[:,i] = data_intensity_ls
            array_id_ls.append(_array_id)
        except:
            sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
            import traceback
            traceback.print_exc()
            sys.stderr.write("array %s, ecotype %s, row %s"%(_array_id, ecotype_id, i))
            
        i += 1
    header = ['probes_id'] + array_id_ls + ['chromosome', 'position']
    del reader
    sys.stderr.write("  Done.\n")
    return data_matrix, probe_id_ls, chr_pos_ls, header

def getProbeIntensityData(input_fname, data_type=numpy.float32, \
    nonIntensityColumnLabelSet = set(['probes_id', 'chromosome', 'position']),\
    array_id = None):
    """
    1. use header[0] to guess the orientation;
    2. call getProbeIntensityDataFromArrayXProbe() or getProbeIntensityDataFromProbeXArray()
    2010-5-23
        add argument array_id to get intensity only for that array.
        add argument nonIntensityColumnLabelSet to find out which columns harbor intensity data.
        Non-intensity columns (probes_id, chromosome, position) are not bound to any particular column.
    2010-3-18
        copied from CNVNormalize.get_input()
    2009-10-28
        switch the default data_type to numpy.float32 to save memory on 64bit machines
    2009-9-28
        add argument data_type to specify data type of data_matrix.
        default is numpy.float (numpy.float could be float32, float64, float128 depending on the architecture).
            numpy.double is also fine.
    2009-5-18
        become classmethod
    """
    sys.stderr.write("Testing which type of intensity data ... ")
    import csv, subprocess
    reader = csv.reader(open(input_fname), delimiter=figureOutDelimiter(input_fname))
    header = next(reader)
    del reader
    if header[0]=='':
        # nothing there. it's Array X Probe format
        return getProbeIntensityDataFromArrayXProbe(input_fname, data_type=data_type, array_id=array_id)
    else:
        return getProbeIntensityDataFromProbeXArray(input_fname, data_type=data_type, \
            nonIntensityColumnLabelSet = nonIntensityColumnLabelSet, array_id=array_id)

class TilingProbeIntensityData(object):
    """
    2010-7-28
    a data structure which makes accessing the intensity data of any array from the big file easy and memory-efficient.
        
    """
    def __init__(self, input_fname = None, min_reciprocal_overlap=None):
        self.input_fname = input_fname
        self.min_reciprocal_overlap = min_reciprocal_overlap
        self.array_id2index = {}
        self.intensity_matrix = []
        self.array_id2gwr = {}
        self.probe_id_ls = None
        self.chr_pos_ls = None
    
    def getIntensityForOneArrayInGWRGivenRBDict(self, array_id, rbDict=None,
        additionalTitle=None, isDataDiscrete=True):
        """
        2010-7-28
        """
        if array_id not in self.array_id2index:	#first check if it's fetched already.		
            data_matrix, probe_id_ls, chr_pos_ls, header = getProbeIntensityData(
                self.input_fname, array_id=array_id)
            
            array_id_ls = header[1:-2]
            array_id_ls = map(int, array_id_ls)
            if len(array_id_ls)==0:	#nothing found for this array in the file
                sys.stderr.write("No data for array %s in %s.\n"%(array_id, self.input_fname))
                return None
            
            if self.probe_id_ls is None:
                self.probe_id_ls = probe_id_ls
            if self.chr_pos_ls is None:
                self.chr_pos_ls = chr_pos_ls
            
            self.array_id2index[array_id] = len(self.array_id2index)
            # data_matrix is in probe X array format.
            self.intensity_matrix.append(data_matrix[:,0])
            
        from SNP import GenomeWideResult, DataObject
        gwr_name = 'array %s probe intensity'%array_id
        if additionalTitle:
            gwr_name += ' for %s'%(additionalTitle)
        
        gwr = GenomeWideResult(name=gwr_name)
        # 2010-3-18 custom
        gwr.array_id = array_id
        #gwr.ecotype_id = array.maternal_ecotype_id
        #gwr.nativename = ecotype_nativename
        
        genome_wide_result_id = id(gwr)
        
        array_index = self.array_id2index.get(array_id)
        
        no_of_rows = len(self.probe_id_ls)
        for i in range(no_of_rows):
            chr_pos = self.chr_pos_ls[i]
            chr, pos = map(int, chr_pos)
            cnvSegmentKey = CNVSegmentBinarySearchTreeKey(chromosome=chr, span_ls=[pos],\
                    min_reciprocal_overlap=self.min_reciprocal_overlap,\
                    isDataDiscrete=isDataDiscrete)
            if rbDict is None or cnvSegmentKey in rbDict:
                probeIntensity = self.intensity_matrix[array_index][i]
                data_obj = DataObject(chromosome=chr, position=pos, value=probeIntensity)
                data_obj.comment = ''
                data_obj.genome_wide_result_name = gwr_name
                data_obj.genome_wide_result_id = genome_wide_result_id
                gwr.add_one_data_obj(data_obj)
        sys.stderr.write(" %s probes. Done.\n"%(len(gwr.data_obj_ls)))
        return gwr

def fetchIntensityInGWAWithinRBDictGivenArrayIDFromTilingIntensity(
    tilingIntensityData, array_id, rbDict, gwr_name=None,\
    min_reciprocal_overlap=0.6, isDataDiscrete=True):
    """
    2010-3-18
        tilingIntensityData is of type SNPData.
        
    """
    sys.stderr.write("Getting intensity data within the chosen segments for array %s ..."%array_id)
    col_index = tilingIntensityData.col_id2col_index.get(array_id)
    if col_index is None:
        sys.stderr.write("Error: No tiling intensity for array %s.\n"%array_id)
        return None
    
    from SNP import GenomeWideResult, DataObject
    
    gwr = GenomeWideResult(name=gwr_name)
    # 2010-3-18 custom
    gwr.array_id = array_id
    #gwr.ecotype_id = array.maternal_ecotype_id
    #gwr.nativename = ecotype_nativename
    
    genome_wide_result_id = id(gwr)
        
    no_of_rows = len(tilingIntensityData.row_id_ls)
    for i in range(no_of_rows):
        chr_pos = tilingIntensityData.row_id_ls[i]
        chromosome, pos = map(int, chr_pos)
        cnvSegmentKey = CNVSegmentBinarySearchTreeKey(
            chromosome=chromosome,
            span_ls=[pos],
            min_reciprocal_overlap=min_reciprocal_overlap,
            isDataDiscrete=isDataDiscrete)
        if cnvSegmentKey in rbDict:
            probeIntensity = tilingIntensityData.data_matrix[i][col_index]
            data_obj = DataObject(chromosome=chromosome,
                position=pos, value=probeIntensity)
            data_obj.comment = ''
            data_obj.genome_wide_result_name = gwr_name
            data_obj.genome_wide_result_id = genome_wide_result_id
            gwr.add_one_data_obj(data_obj)
    sys.stderr.write(" %s probes. Done.\n"%(len(gwr.data_obj_ls)))
    return gwr


class ArrayXProbeFileWrapper(object):
    """
    2010-6-3
        a wrapper around CNV input file, which is in array X probe format.
        It will read 3 lines of headers first and then let users to access file through self.reader.
    """
    def __init__(self, input_fname):
        """
        """
        self.input_fname = input_fname
        import csv
        from palos import figureOutDelimiter
        self.reader = csv.reader(open(self.input_fname),
            delimiter=figureOutDelimiter(self.input_fname))
        self.probe_id_ls, self.chr_pos_ls = self.getHeader(self.reader)
    
    @classmethod
    def getHeader(cls, reader):
        sys.stderr.write("Getting probe id, chromosome, pos info ...")
        probe_id_ls = next(reader)[2:]
        probe_id_ls = map(int, probe_id_ls)
        chr_ls = next(reader)[2:]
        pos_ls = next(reader)[2:]
        chr_pos_ls = zip(chr_ls, pos_ls)
        sys.stderr.write(".\n")
        return probe_id_ls, chr_pos_ls

def get_chr2start_stop_index(chr_pos_ls):
    """
    2010-6-5
        get the boundary of all chromosomes based on chr_pos_ls
    """
    sys.stderr.write("Getting chr2start_stop_index ...")
    no_of_probes = len(chr_pos_ls)
    chr2start_stop_index = {}
    old_chr = None
    for i in range(no_of_probes):
        chromosome = chr_pos_ls[i][0]
        if chromosome not in chr2start_stop_index:
            chr2start_stop_index[chromosome] = [i]
        if old_chr is not None and chromosome!=old_chr:
            chr2start_stop_index[old_chr].append(i-1)
        
        old_chr = chromosome
    #append the the stop index to the last chromosome
    chr2start_stop_index[old_chr].append(i)
    sys.stderr.write("Done.\n")
    return chr2start_stop_index

def readQuanLongPECoverageIntoGWR(input_fname, additionalTitle=None,
    windowSize=100, chr=None, start=None, stop=None):
    """
    This function reads 100bp coverage data from files given by Quan and turns it into instance of GenomeWideResult.
        example file: "variation/data/CNV/QuanLongPE/coverage4Yu/algustrum.8230.Chr1.coverage"
        format is two-column, tab-delimited. First column is position/100. 2nd column is coverage.:
            1       0.0
            2       0.0
            3       2.84
        
        the coverage data is split into different chromosomes. so argument chr is not used here.
    """
    from SNP import GenomeWideResult, DataObject
    import os, sys, re, csv
    sys.stderr.write("Reading Quan's coverage data from %s ... "%(input_fname))
    chrPattern = re.compile(r'.*Chr(\d+).coverage')
    chrPatternSearchResult = chrPattern.search(input_fname)
    if chrPatternSearchResult:
        chr = int(chrPatternSearchResult.group(1))
    else:
        sys.stderr.write("Can't parse chromosome out of %s. aborted.\n"%input_fname)
    reader = csv.reader(open(input_fname), delimiter='\t')
    gwr_name = 'log10(coverage) %s'%os.path.basename(input_fname)
    if additionalTitle:
        gwr_name += " "+ additionalTitle
    
    coverage_gwr = GenomeWideResult(name=gwr_name)
    
    genome_wide_result_id = id(coverage_gwr)
    
    for row in reader:
        position, coverage = row[:2]
        position = int(position)*windowSize-windowSize/2
        #coverage is caculated in 100-base window.
        if start is not None and position<start:
            continue
        if stop is not None and position>stop:
            continue
        coverage = float(coverage)
        if coverage<=0.0:
            coverage = -3
        else:
            coverage = math.log10(coverage)
        data_obj = DataObject(chromosome=chr, position=position, value=coverage)
        data_obj.comment = ''
        data_obj.genome_wide_result_name = gwr_name
        data_obj.genome_wide_result_id = genome_wide_result_id
        coverage_gwr.add_one_data_obj(data_obj)
    sys.stderr.write("Done.\n")
    return coverage_gwr


def findCorrespondenceBetweenTwoCNVRBDict(rbDict1=None, rbDict2=None, isDataDiscrete=True):
    """
    2013.1.27
    """
    dc1Length = len(rbDict1)
    dc2Length = len(rbDict2)
    sys.stderr.write("Finding correspondence between two CNV RB dictionaries, %s and %s nodes in rbDict1, rbDict2 respectively ... "%\
        (dc1Length, dc2Length))
    nodePairList = []
    #each cell is a tuple of (node1, node2, overlap)
    #to detect any overlap
    compareIns = CNVCompare(min_reciprocal_overlap=0.0000001)
    rb_dict2_nodes_matching_rb_dict1_set = set()
    for rb_dict1_node in rbDict1:
        targetNodeLs = []
        rbDict2.findNodes(rb_dict1_node.key, node_ls=targetNodeLs, compareIns=compareIns)
        if targetNodeLs:
            for rb_dict2_node in targetNodeLs:
                overlapData = get_overlap_ratio(rb_dict1_node.key.span_ls,
                    rb_dict2_node.key.span_ls, isDataDiscrete=isDataDiscrete)
                overlapFraction = overlapData.overlapFraction
                nodePairList.append((rb_dict1_node, rb_dict2_node, overlapFraction))
                rb_dict2_nodes_matching_rb_dict1_set.add(rb_dict2_node)
        else:
            nodePairList.append((rb_dict1_node, None, None))
    sys.stderr.write(" %s input2 nodes matched. "%(len(rb_dict2_nodes_matching_rb_dict1_set)))
    
    no_of_unmatched_input2_nodes_matched_to_input1 = 0
    for rb_dict2_node in rbDict2:
        if rb_dict2_node not in rb_dict2_nodes_matching_rb_dict1_set:
            targetNodeLs = []
            rbDict1.findNodes(rb_dict2_node.key, node_ls=targetNodeLs, compareIns=compareIns)
            if targetNodeLs:	#this should not happen
                sys.stderr.write("Warning (should not happen): rbDict2-only node (key=%s, value=%s) was found to have some matches in rbDict1.\n"%\
                    (str(rb_dict2_node.key), str(rb_dict2_node.value[0])))
                no_of_unmatched_input2_nodes_matched_to_input1 += 1
                
                for rb_dict1_node in targetNodeLs:
                    overlapData = get_overlap_ratio(rb_dict1_node.key.span_ls,
                        rb_dict2_node.key.span_ls, isDataDiscrete=isDataDiscrete)
                    overlapFraction = overlapData.overlapFraction
                    nodePairList.append((rb_dict1_node, rb_dict2_node, overlapFraction))
            else:
                nodePairList.append((None, rb_dict2_node, None))
    sys.stderr.write(" %s un-matched input2 nodes have matches from input1. found %s corresponding pairs.\n"%\
        (no_of_unmatched_input2_nodes_matched_to_input1, len(nodePairList)))
    return nodePairList


# test some functions within this module.
if __name__ == "__main__":
    import os, sys, math
    
    if len(sys.argv)>=2 and sys.argv[1]=='-b':
        # 2010-6-17 enter debug mode "CNV.py -b"
        import pdb
        pdb.set_trace()
        debug = True
    else:
        debug = False
    
    cnv_ls = [[1, (2323,2600)], [2,(50000,)], [3,(43214,78788)], [5,(150,500)], \
              [5,(500,950)], [5, (43241, 43242)]]
    no_of_cnvs = len(cnv_ls)
    min_reciprocal_overlap = 0.6
    
    #from BinarySearchTree import binary_tree
    #tree = binary_tree()
    from palos.algorithm.RBTree import RBDict
    # 2010-1-26 binary_tree and RBDict are swappable. but RBDict is more efficient (balanced).
    tree = RBDict(cmpfn=leftWithinRightAlsoEqualCmp)
    # 2010-1-28 use the custom cmpfn if you want the case that left within right is regarded as equal as well.
    tree = RBDict(cmpfn=rightWithinLeftAlsoEqualCmp)
    
    for cnv in cnv_ls:
        segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=cnv[0],
            span_ls=cnv[1], min_reciprocal_overlap=min_reciprocal_overlap)
        tree[segmentKey] = cnv
    
    print("Binary Tree Test\n")
    print("Node Count: %d" % len(tree))
    print("Depth: %d" % tree.depth())
    print("Optimum Depth: %f (%d) (%f%% depth efficiency)" % (
        tree.optimumdepth(), math.ceil(tree.optimumdepth()),
        math.ceil(tree.optimumdepth()) / tree.depth()))
    
    print("Efficiency: %f%% (total possible used: %d, total wasted: %d)" % (
        tree.efficiency() * 100,
        len(tree) / tree.efficiency(),
        (len(tree) / tree.efficiency()) - len(tree)))
    """
    print("Min: %s" % repr(tree.min()))
    print("Max: %s" % repr(tree.max()))
    
    print("List of Layers:\n\t" + repr(tree.listlayers()) + "\n")
    print("\"Recursive\" List:\n\t" + repr(tree.listrecursive()) + "\n")
    print("List of Keys:\n\t" + repr(tree.listkeys()) + "\n")
    print("List of Data:\n\t" + repr(tree.listdata()) + "\n")
    print("List of Nodes:\n\t" + repr(tree.listnodes()) + "\n")
    print("Dictionary:\n\t" + repr(tree.dict()) + "\n")
    print("Formatted Tree:\n" + tree.formattree() + "\n")
    print("Formatted Tree (Root in Middle):\n" + tree.formattreemiddle() + "\n")
    """
    test_cnv_ls = [	[2,(50000,)], [3,(43214,43219)], [3,(43214,78788)], \
        [5, (43242,)], [5,(144,566)], [5, (144, 1000)], [5, (50000, 70000)]]
    for test_cnv in test_cnv_ls:
        segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=test_cnv[0], span_ls=test_cnv[1], \
            min_reciprocal_overlap=min_reciprocal_overlap)
        print("segmentKey", segmentKey)
        if segmentKey in tree:
            targetSegment = tree.get(segmentKey)
            #if targetSegment:
            print("\tIn tree with target", targetSegment)
            node_ls = []
            tree.findNodes(segmentKey, node_ls)
            print("\tfindNodes()", [node.value for node in node_ls])
        else:
            print("\tNot in tree")
    
    
    """
    for i in range(no_of_cnvs):
        for j in range(i+1, no_of_cnvs):
            cnv1 = cnv_ls[i]
            cnv2 = cnv_ls[j]
            segmentKey1 = CNVSegmentBinarySearchTreeKey(chromosome=cnv1[0], span_ls=cnv1[1], \
                    min_reciprocal_overlap=min_reciprocal_overlap)
            segmentKey2 = CNVSegmentBinarySearchTreeKey(chromosome=cnv2[0], span_ls=cnv2[1], \
                    min_reciprocal_overlap=min_reciprocal_overlap)
            print segmentKey1, "vs", segmentKey2
            print(">", segmentKey1>segmentKey2)
            print(">=", segmentKey1>=segmentKey2)
            print("<", segmentKey1<segmentKey2)
            print("<=", segmentKey1<=segmentKey2)
            print("==", segmentKey1==segmentKey2)
    """
